"""
Flitter generic rendering
"""


class Renderer:
    """
    Renderers should implement this API
    """
    def __init__(self, **kwargs):
        """
        `kwargs` will contain the engine global definitions. You will likely
        want to create most of the renderer state on first call of `update()`
        when you have access to the top-level node attributes.
        """
        pass

    async def purge(self):
        """
        `purge()`` should wipe any local state ready for a complete redefinition.
        """
        raise NotImplementedError()

    async def destroy(self):
        """
        `destroy()`` should tear down the renderer itself.
        """
        raise NotImplementedError()

    async def update(self, engine, node, **kwargs):
        """
        `update()`` is called every frame with the node to render and the engine
        global definitions (including `time` and `beat`).
        """
        raise NotImplementedError()



"""
Flitter window management
"""

import array
import asyncio
import ctypes
import os
import sys

import glfw
from loguru import logger
from mako.template import Template
import moderngl

from ...clock import system_clock
from .glsl import TemplateLoader
from .target import RenderTarget, COLOR_FORMATS
from ...model import Vector, true, false
from ...plugins import get_plugin


def set_uniform_vector(uniform, vector):
    dtype = {'f': float, 'd': float, 'i': int, 'I': int}[uniform.fmt[-1]]
    n, m = uniform.array_length, uniform.dimension
    if (value := vector.match(n * m, dtype)) is not None:
        if m == 1:
            uniform.value = value if n == 1 else list(value)
        elif n == 1:
            uniform.value = tuple(value)
        else:
            uniform.value = [tuple(value[i*m:(i+1)*m]) for i in range(n)]
        return True
    elif n > 1 and m > 1 and (value := vector.match(m, dtype)) is not None:
        uniform.value = [tuple(value) for i in range(n)]
        return True
    return False


RESERVED_NAMES = {'fragment', 'vertex', 'loop', 'context', 'UNDEFINED'}
DEFAULT_LINEAR = False
DEFAULT_COLORBITS = 16
PUSHED = Vector.symbol('pushed')
RELEASED = Vector.symbol('released')
PUSHED_BEAT = Vector.symbol('pushed').concat(Vector.symbol('beat'))
RELEASED_BEAT = Vector.symbol('released').concat(Vector.symbol('beat'))

GLFUNCTYPE = ctypes.WINFUNCTYPE if sys.platform == 'win32' else ctypes.CFUNCTYPE


class WindowNode:
    def __init__(self, glctx):
        self.glctx = glctx
        self.children = []
        self.node_id = None
        self.width = None
        self.height = None
        self.tags = set()
        self.hidden = False

    @property
    def name(self):
        return '#'.join((self.__class__.__name__.lower(), *self.tags))

    @property
    def texture(self):
        raise NotImplementedError()

    @property
    def array(self):
        raise NotImplementedError()

    @property
    def child_textures(self):
        textures = {}
        i = 0
        for child in self.children:
            if child.texture is not None and not child.hidden:
                textures[f'texture{i}'] = child.texture
                i += 1
        return textures

    async def destroy(self):
        await self.purge()
        result = self.release()
        if asyncio.coroutines.iscoroutine(result):
            await result
        self.glctx = None

    async def purge(self):
        while self.children:
            await self.children.pop().destroy()

    def release(self):
        pass

    async def update(self, engine, node, references, *, default_size=(512, 512), **kwargs):
        self.node_id = node.get('id', 1, str)
        if self.node_id is not None:
            references[self.node_id] = self
        self.hidden = node.get('hidden', 1, bool, False)
        resized = False
        width, height = node.get('size', 2, int, default_size)
        if width != self.width or height != self.height:
            self.width = width
            self.height = height
            resized = True
        self.tags = node.tags
        result = self.create(engine, node, resized, default_size=(self.width, self.height), **kwargs)
        if asyncio.coroutines.iscoroutine(result):
            await result
        await self.descend(engine, node, references, default_size=(self.width, self.height), **kwargs)
        result = self.render(node, references, **kwargs)
        if asyncio.coroutines.iscoroutine(result):
            await result
        for child in self.children:
            result = child.parent_finished()
            if asyncio.coroutines.iscoroutine(result):
                await result

    def similar_to(self, node):
        return node.tags == self.tags

    def create(self, engine, node, resized, **kwargs):
        pass

    async def descend(self, engine, node, references, **kwargs):
        existing = self.children
        updated = []
        for child in node.children:
            result = self.handle_node(engine, child, **kwargs)
            if asyncio.coroutines.iscoroutine(result):
                result = await result
            if result:
                continue
            cls = get_plugin('flitter.render.window', child.kind)
            if cls is not None:
                for i, scene_node in enumerate(existing):
                    if type(scene_node) is cls and scene_node.similar_to(child):
                        scene_node = existing.pop(i)
                        break
                else:
                    scene_node = cls(self.glctx)
                await scene_node.update(engine, child, references, **kwargs)
                updated.append(scene_node)
        while existing:
            await existing.pop().destroy()
        self.children = updated

    def render(self, node, references, **kwargs):
        raise NotImplementedError()

    def parent_finished(self):
        pass

    def handle_node(self, engine, node, **kwargs):
        return False


class Reference(WindowNode):
    def __init__(self, glctx):
        super().__init__(glctx)
        self._reference = None

    def release(self):
        self._reference = None

    @property
    def texture(self):
        return self._reference.texture if self._reference is not None else None

    @property
    def array(self):
        return self._reference.array if self._reference is not None else None

    async def update(self, engine, node, references, **kwargs):
        node_id = node.get('id', 1, str)
        self._reference = references.get(node_id) if node_id else None


class ProgramNode(WindowNode):
    DEFAULT_VERTEX_SOURCE = TemplateLoader.get_template('default.vert')
    DEFAULT_FRAGMENT_SOURCE = TemplateLoader.get_template('default.frag')

    def __init__(self, glctx):
        super().__init__(glctx)
        self._program = None
        self._rectangle = None
        self._vertex_source = None
        self._fragment_source = None
        self._target = None
        self._retain_target = False

    def release(self):
        if self._target is not None:
            self._target.release()
        self._target = None
        self._program = None
        self._rectangle = None
        self._vertex_source = None
        self._fragment_source = None

    def parent_finished(self):
        if not self._retain_target and self._target is not None:
            self._target.release()
            self._target = None

    @property
    def size(self):
        return self.width, self.height

    @property
    def texture(self):
        return self._target.texture if self._target is not None else None

    @property
    def array(self):
        return self._target.array if self._target is not None else None

    def get_vertex_source(self, node, **defaults):
        vertex = Template(node.get('vertex', 1, str), lookup=TemplateLoader) if 'vertex' in node else self.DEFAULT_VERTEX_SOURCE
        names = dict(defaults)
        for name, value in node.items():
            if name not in RESERVED_NAMES:
                names[name] = value
        names['child_textures'] = list(self.child_textures)
        names['HEADER'] = self.glctx.extra['HEADER']
        return vertex.render(**names).strip()

    def get_fragment_source(self, node, **defaults):
        fragment = Template(node.get('fragment', 1, str), lookup=TemplateLoader) if 'fragment' in node else self.DEFAULT_FRAGMENT_SOURCE
        names = dict(defaults)
        for name, value in node.items():
            if name not in RESERVED_NAMES:
                names[name] = value
        names['child_textures'] = list(self.child_textures)
        names['HEADER'] = self.glctx.extra['HEADER']
        return fragment.render(**names).strip()

    def compile(self, node, **kwargs):
        vertex_source = self.get_vertex_source(node, **kwargs)
        fragment_source = self.get_fragment_source(node, **kwargs)
        if vertex_source != self._vertex_source or fragment_source != self._fragment_source:
            self._vertex_source = vertex_source
            self._fragment_source = fragment_source
            self._program = None
            self._rectangle = None
            try:
                start = system_clock()
                self._program = self.glctx.program(vertex_shader=self._vertex_source, fragment_shader=self._fragment_source)
                vertices = self.glctx.buffer(array.array('f', [-1, 1, -1, -1, 1, 1, 1, -1]))
                self._rectangle = self.glctx.vertex_array(self._program, [(vertices, '2f', 'position')], mode=moderngl.TRIANGLE_STRIP)
            except Exception as exc:
                error = str(exc).strip().split('\n')
                logger.error("{} GL program compile failed: {}", self.name, error[-1])
                source = self._fragment_source if 'fragment_shader' in error else self._vertex_source
                source = '\n'.join(f'{i+1:3d}|{line}' for i, line in enumerate(source.split('\n')))
                logger.trace("Failing source:\n{}", source)
            else:
                end = system_clock()
                logger.debug("{} GL program compiled in {:.1f}ms", self.name, 1000 * (end - start))

    def render(self, node, references, *, srgb=False, colorbits=None, composite='over', passes=1, downsample=2, downsample_passes=(),
               border=None, repeat=None, **kwargs):
        composite = node.get('composite', 1, str, node.get('blend', 1, str, composite))
        passes = max(1, node.get('passes', 1, int, passes))
        downsample = max(1, node.get('downsample', 1, int, downsample))
        downsample_passes = set(node.get('downsample_passes', 0, int, downsample_passes))
        colorbits = node.get('colorbits', 1, int, self.glctx.extra['colorbits'] if colorbits is None else colorbits)
        self.compile(node, passes=passes, composite=composite, **kwargs)
        if self._rectangle is None:
            return
        border = node.get('border', 4, float, border)
        repeat = node.get('repeat', 2, bool, repeat)
        if border is not None:
            sampler_args = {'border_color': tuple(border)}
        elif repeat is not None:
            sampler_args = {'repeat_x': repeat[0], 'repeat_y': repeat[1]}
        else:
            sampler_args = {'border_color': (0, 0, 0, 0)}
        child_textures = self.child_textures
        samplers = []
        self.glctx.extra['zero'].use(0)
        unit = 1
        pass_member = None
        last_member = None
        first_member = None
        size_member = None
        downsample_member = None
        self._retain_target = self.node_id is not None
        for name in self._program:
            member = self._program[name]
            if isinstance(member, moderngl.Uniform):
                if name == 'last':
                    self._retain_target = True
                    last_member = member
                    last_member.value = 0
                elif name == 'first':
                    first_member = member
                    first_member.value = 0
                elif name == 'pass':
                    pass_member = member
                elif name == 'size':
                    size_member = member
                elif name == 'downsample':
                    downsample_member = member
                elif name in child_textures:
                    texture = child_textures[name]
                    if texture is not None:
                        sampler = self.glctx.sampler(texture=child_textures[name], **sampler_args)
                        sampler.use(location=unit)
                        samplers.append(sampler)
                        member.value = unit
                        unit += 1
                    else:
                        member.value = 0
                elif name in node and set_uniform_vector(member, node[name]):
                    pass
                elif name in kwargs and set_uniform_vector(member, Vector.coerce(kwargs[name])):
                    pass
                elif name in 'alpha':
                    set_uniform_vector(member, true)
                else:
                    logger.trace("Unbound uniform: {}", name)
                    set_uniform_vector(member, false)
        self.glctx.enable(moderngl.BLEND)
        self.glctx.blend_func = moderngl.ONE, moderngl.ONE_MINUS_SRC_ALPHA
        last = self._target
        last_sampler = None
        first_sampler = None
        first = None
        for pass_number in range(passes):
            if pass_member is not None:
                pass_member.value = pass_number
            if last_member is not None and last is not None:
                last_sampler = self.glctx.sampler(texture=last.texture, **sampler_args)
                last_sampler.use(location=unit)
                last_member.value = unit
            if first_member is not None and pass_number == 1:
                first_sampler = self.glctx.sampler(texture=first.texture, **sampler_args)
                first_sampler.use(location=unit+1)
                first_member.value = unit+1
            if pass_number in downsample_passes:
                target = RenderTarget.get(self.glctx, self.width//downsample, self.height//downsample, colorbits, srgb=srgb)
                if downsample_member is not None:
                    downsample_member.value = downsample
            else:
                target = RenderTarget.get(self.glctx, self.width, self.height, colorbits, srgb=srgb)
                if downsample_member is not None:
                    downsample_member.value = 1
            if size_member is not None:
                size_member.value = target.size
            with target:
                target.clear()
                self._rectangle.render()
            if first_member is not None and pass_number == 0:
                first = target
            if last is not None and last is not first:
                last.release()
            last = target
            if last_sampler is not None:
                last_sampler.clear()
        if first_sampler is not None:
            first_sampler.clear()
        self._target = last
        if first is not None and first is not last:
            first.release()
        for sampler in samplers:
            sampler.clear()
        self.glctx.disable(moderngl.BLEND)


class GLFWLoader:
    FUNCTIONS = {}
    ALLOW_MISSING = {'glPrimitiveRestartIndex'}

    @staticmethod
    def load_opengl_function(name):
        address = glfw.get_proc_address(name)
        if address is not None:
            return address
        if name in GLFWLoader.ALLOW_MISSING:
            return 0
        if (function := getattr(GLFWLoader, 'shim_' + name, None)) is not None:
            logger.debug("Use shim for missing GL function: {}", name)
            return ctypes.cast(function, ctypes.c_void_p).value
        if (function := GLFWLoader.FUNCTIONS.get(name)) is not None:
            return ctypes.cast(function, ctypes.c_void_p).value
        logger.trace("Request for missing GL function: {}", name)

        def missing(name):
            def function():
                logger.error("Call to missing GL function: {}", name)
                os._exit(1)
            return function

        function = GLFUNCTYPE(None)(missing(name))
        GLFWLoader.FUNCTIONS[name] = function
        return ctypes.cast(function, ctypes.c_void_p).value

    @staticmethod
    def get_function(name, *signature):
        function = GLFWLoader.FUNCTIONS.get(name)
        if function is None:
            function = ctypes.cast(glfw.get_proc_address(name), GLFUNCTYPE(*signature))
            GLFWLoader.FUNCTIONS[name] = function
        return function

    @GLFUNCTYPE(None, ctypes.c_uint, ctypes.c_uint)
    @staticmethod
    def shim_glClampColor(target, clamp):
        pass

    @GLFUNCTYPE(None, ctypes.c_double)
    @staticmethod
    def shim_glClearDepth(depth):
        glClearDepthf = GLFWLoader.get_function('glClearDepthf', None, ctypes.c_float)
        glClearDepthf(depth)

    @GLFUNCTYPE(None, ctypes.c_uint)
    @staticmethod
    def shim_glDrawBuffer(buf):
        glDrawBuffers = GLFWLoader.get_function('glDrawBuffers', None, ctypes.c_uint, ctypes.POINTER(ctypes.c_uint))
        glDrawBuffers(1, ctypes.byref(ctypes.c_uint(buf)))

    @staticmethod
    def release():
        pass


class Window(ProgramNode):
    WINDOW_FRAGMENT_SOURCE = TemplateLoader.get_template('window.frag')
    Windows = []

    def __init__(self, screen=0, fullscreen=False, vsync=False, offscreen=False, **kwargs):
        super().__init__(None)
        self.engine = None
        self.window = None
        self.default_screen = screen
        self.default_fullscreen = fullscreen
        self.default_vsync = vsync
        self._deferred_fullscreen = False
        self._visible = not offscreen
        self._screen = None
        self._fullscreen = None
        self._resizable = None
        self._title = None
        self._beat = None

    def release(self):
        super().release()
        if self.glctx is not None:
            self.glctx.finish()
            RenderTarget.empty_pool(self.glctx)
            self.glctx.extra.clear()
            if count := self.glctx.gc():
                logger.trace("Window release collected {} OpenGL objects", count)
            self.glctx.release()
            self.glctx = None
        if self.window is not None:
            glfw.destroy_window(self.window)
            Window.Windows.remove(self)
            if not Window.Windows:
                glfw.terminate()
            logger.debug("{} closed", self.name)
            self.window = None
        self.engine = None

    def create(self, engine, node, resized, *, opengl_es=False, **kwargs):
        super().create(engine, node, resized, opengl_es=opengl_es, **kwargs)
        self._keys = {}
        self._pointer_state = None
        new_window = False
        screen = node.get('screen', 1, int, self.default_screen)
        fullscreen = node.get('fullscreen', 1, bool, self.default_fullscreen) if self._visible else False
        resizable = node.get('resizable', 1, bool, True) if self._visible else False
        title = node.get('title', 1, str, "Flitter")
        if self.window is None:
            self.engine = engine
            if not Window.Windows:
                ok = glfw.init()
                if not ok:
                    raise RuntimeError("Unable to initialize GLFW")
                logger.debug("GLFW version: {}", glfw.get_version_string().decode('utf8'))
            glfw.window_hint(glfw.CONTEXT_CREATION_API, glfw.EGL_CONTEXT_API if opengl_es else glfw.NATIVE_CONTEXT_API)
            glfw.window_hint(glfw.CLIENT_API, glfw.OPENGL_ES_API if opengl_es else glfw.OPENGL_API)
            glfw.window_hint(glfw.CONTEXT_VERSION_MAJOR, 3)
            glfw.window_hint(glfw.CONTEXT_VERSION_MINOR, 0 if opengl_es else 3)
            glfw.window_hint(glfw.OPENGL_PROFILE, glfw.OPENGL_ANY_PROFILE if opengl_es else glfw.OPENGL_CORE_PROFILE)
            glfw.window_hint(glfw.OPENGL_FORWARD_COMPAT, glfw.TRUE)
            glfw.window_hint(glfw.SRGB_CAPABLE, glfw.FALSE)
            glfw.window_hint(glfw.SAMPLES, 0)
            glfw.window_hint(glfw.VISIBLE, glfw.TRUE if self._visible else glfw.FALSE)
            if self._visible:
                glfw.window_hint(glfw.DOUBLEBUFFER, glfw.TRUE)
                glfw.window_hint(glfw.AUTO_ICONIFY, glfw.FALSE)
                glfw.window_hint(glfw.CENTER_CURSOR, glfw.FALSE)
                glfw.window_hint(glfw.SCALE_TO_MONITOR, glfw.TRUE)
            self.window = glfw.create_window(self.width, self.height, title, None, Window.Windows[0].window if Window.Windows else None)
            glfw.set_key_callback(self.window, self.key_callback)
            glfw.set_cursor_pos_callback(self.window, self.pointer_movement_callback)
            glfw.set_mouse_button_callback(self.window, self.pointer_button_callback)
            self._title = title
            Window.Windows.append(self)
            new_window = True
            if self._visible and hasattr(glfw, 'get_cocoa_window'):
                try:
                    import objc
                    import Cocoa
                    nswindow = objc.objc_object(c_void_p=ctypes.c_void_p(glfw.get_cocoa_window(self.window)))
                    nswindow.setColorSpace_(Cocoa.NSColorSpace.sRGBColorSpace())
                    logger.debug("Set macOS window to sRGB colorspace")
                except ImportError:
                    pass
                except Exception:
                    logger.exception("Failed to set macOS window to sRGB colorspace")
        if self._visible and resizable != self._resizable:
            glfw.set_window_attrib(self.window, glfw.RESIZABLE, glfw.TRUE if resizable else glfw.FALSE)
            self._resizable = resizable
        if self._visible and (resized or screen != self._screen or fullscreen != self._fullscreen):
            monitors = glfw.get_monitors()
            monitor = monitors[screen] if screen < len(monitors) else monitors[0]
            mx, my, mw, mh = glfw.get_monitor_workarea(monitor)
            width, height = self.width, self.height
            if not fullscreen:
                glfw.set_window_aspect_ratio(self.window, self.width, self.height)
            while width > mw * 0.95 or height > mh * 0.95:
                width = width * 2 // 3
                height = height * 2 // 3
            if self._fullscreen and not fullscreen or screen != self._screen:
                glfw.set_window_monitor(self.window, None, 0, 0, width, height, glfw.DONT_CARE)
            if new_window or (self._fullscreen and not fullscreen) or screen != self._screen:
                glfw.set_window_pos(self.window, mx + (mw - width) // 2, my + (mh - height) // 2)
            if fullscreen and (not self._fullscreen or screen != self._screen):
                mode = glfw.get_video_mode(monitor)
                glfw.set_window_monitor(self.window, monitor, 0, 0, mode.size.width, mode.size.height, mode.refresh_rate)
            if not fullscreen:
                glfw.set_window_size(self.window, width, height)
            self._screen = screen
            self._fullscreen = fullscreen
        if title != self._title:
            glfw.set_window_title(self.window, title)
            self._title = title
        glfw.make_context_current(self.window)
        if new_window:
            if self._visible:
                logger.debug("{} opened on screen {}", self.name, screen)
            else:
                logger.debug("{} opened off-screen", self.name)
            self.glctx = moderngl.create_context(require=300 if opengl_es else 330, context=GLFWLoader)
            self.glctx.gc_mode = 'context_gc'
            logger.debug("OpenGL info: {GL_RENDERER} {GL_VERSION}", **self.glctx.info)
            logger.trace("{!r}", self.glctx.info)
            zero = self.glctx.texture((1, 1), 4, dtype='f1')
            zero.write(bytes([0, 0, 0, 0]))
            if opengl_es:
                header = "#version 300 es\nprecision highp float;\nprecision highp int;\n"
            else:
                header = "#version 330\n"
            self.glctx.extra = {'zero': zero, 'HEADER': header}
            vertex_source = self.DEFAULT_VERTEX_SOURCE.render(HEADER=header)
            fragment_source = self.WINDOW_FRAGMENT_SOURCE.render(HEADER=header)
            self._window_program = self.glctx.program(vertex_shader=vertex_source, fragment_shader=fragment_source)
            vertices = self.glctx.buffer(array.array('f', [-1, 1, -1, -1, 1, 1, 1, -1]))
            self._window_rectangle = self.glctx.vertex_array(self._window_program, [(vertices, '2f', 'position')], mode=moderngl.TRIANGLE_STRIP)
        if self._visible:
            self.recalculate_viewport(new_window)
        colorbits = node.get('colorbits', 1, int, DEFAULT_COLORBITS)
        if colorbits not in COLOR_FORMATS:
            colorbits = DEFAULT_COLORBITS
        self.glctx.extra['colorbits'] = colorbits

    def key_callback(self, window, key, scancode, action, mods):
        if key in self._keys:
            state = self._keys[key]
            if action == glfw.PRESS:
                self.engine.state[state] = 1
                self.engine.state[state.concat(PUSHED)] = 1
                self.engine.state[state.concat(RELEASED)] = 0
                self.engine.state[state.concat(PUSHED_BEAT)] = self._beat
            elif action == glfw.RELEASE:
                self.engine.state[state] = 0
                self.engine.state[state.concat(PUSHED)] = 0
                self.engine.state[state.concat(RELEASED)] = 1
                self.engine.state[state.concat(RELEASED_BEAT)] = self._beat
        elif key == glfw.KEY_LEFT:
            if action == glfw.RELEASE:
                self.engine.previous_page()
        elif key == glfw.KEY_RIGHT:
            if action == glfw.RELEASE:
                self.engine.next_page()

    def pointer_movement_callback(self, window, x, y):
        if self._pointer_state is not None and self.window is not None:
            width, height = glfw.get_window_size(self.window)
            if 0 <= x <= width and 0 <= y <= height:
                self.engine.state[self._pointer_state] = x/width, y/height
            else:
                self.engine.state[self._pointer_state] = None

    def pointer_button_callback(self, window, button, action, mods):
        if self._pointer_state is not None and self.window is not None:
            if action == glfw.PRESS:
                self.engine.state[self._pointer_state.concat(Vector(button))] = 1
            elif action == glfw.RELEASE:
                self.engine.state[self._pointer_state.concat(Vector(button))] = 0

    def handle_node(self, engine, node, **kwargs):
        if node.kind == 'key':
            if 'state' in node:
                key_name = node.get('name', 1, str)
                if key_name:
                    key_constant = 'KEY_' + key_name.upper()
                    if hasattr(glfw, key_constant):
                        key = getattr(glfw, key_constant)
                        state = node['state']
                        self._keys[key] = state
                        if state not in engine.state:
                            if self.window is not None and glfw.get_key(self.window, key) == glfw.PRESS:
                                engine.state[state] = 1
                                engine.state[state.concat(PUSHED)] = 1
                                engine.state[state.concat(RELEASED)] = 0
                            else:
                                engine.state[state] = 0
                                engine.state[state.concat(PUSHED)] = 0
                                engine.state[state.concat(RELEASED)] = 1
            return True
        elif node.kind == 'pointer':
            if 'state' in node:
                self._pointer_state = node['state']
            return True
        return super().handle_node(engine, node, **kwargs)

    def recalculate_viewport(self, force=False):
        aspect_ratio = self.width / self.height
        width, height = glfw.get_framebuffer_size(self.window)
        if width / height > aspect_ratio:
            view_width = int(height * aspect_ratio)
            viewport = ((width - view_width) // 2, 0, view_width, height)
        else:
            view_height = int(width / aspect_ratio)
            viewport = (0, (height - view_height) // 2, width, view_height)
        if force or viewport != self.glctx.screen.viewport:
            self.glctx.screen.viewport = viewport
            width, height = glfw.get_window_size(self.window)
            if viewport != (0, 0, width, height):
                logger.debug("{0} resized to {1}x{2} (viewport {5}x{6} x={3} y={4})", self.name, width, height, *viewport)
            else:
                logger.debug("{} resized to {}x{}", self.name, width, height)

    def render(self, node, references, beat=None, **kwargs):
        if self._visible or self.node_id is not None:
            super().render(node, references, **kwargs)
        if self._visible:
            self.glctx.screen.use()
            self.glctx.screen.clear(0, 0, 0, 1)
            self._target.texture.use(1)
            self._window_program['target'] = 1
            self._window_rectangle.render()
            vsync = node.get('vsync', 1, bool, self.default_vsync)
            glfw.swap_interval(1 if vsync else 0)
            glfw.swap_buffers(self.window)
        else:
            self.glctx.finish()
        self._beat = beat
        glfw.poll_events()
        RenderTarget.empty_pool(self.glctx, 15)
        self.glctx.gc()

    def make_secondary_texture(self):
        width, height = self.window.get_framebuffer_size()
        return self.glctx.texture((width, height), 4)


class Offscreen(Window):
    def __init__(self, offscreen=False, **kwargs):
        super().__init__(offscreen=True, **kwargs)




cdef extern from *:
    """
#include "Python.h"

#if PY_MINOR_VERSION >= 13
    static inline double perf_counter() {
        PyTime_t result;
        PyTime_PerfCounter(&result);
        return PyTime_AsSecondsDouble(result);
    }
#else
    static inline double perf_counter() {
        return _PyTime_AsSecondsDouble(_PyTime_GetPerfCounter());
    }
#endif
    """

    cdef double perf_counter() noexcept nogil



"""
Common Flitter initialisation
"""

import logging
import sys

from loguru import logger

try:
    import pyximport
    pyximport.install(language_level=3)
except ImportError:
    pass

try:
    from setproctitle import setproctitle
except ImportError:
    def setproctitle(title):
        pass


__version__ = "1.0.0b22"

LOGGING_LEVEL = "SUCCESS"
LOGGING_FORMAT = "{time:HH:mm:ss.SSS} {extra[shortname]:25} | <level>{level}: {message}</level>"


class LoguruInterceptHandler(logging.Handler):
    @classmethod
    def install(cls):
        handler = cls()
        logging.basicConfig(handlers=[handler], level=0)
        return handler

    def uninstall(self):
        logging.getLogger().removeHandler(self)

    def emit(self, record):
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno
        frame, depth = logging.currentframe().f_back, 1
        while frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1
        logger.opt(depth=depth, exception=record.exc_info).log(level, record.getMessage())


def shorten_name(pkg_name, n=25):
    if len(pkg_name) <= n:
        return pkg_name
    over = len(pkg_name) - n
    parts = pkg_name.split('.')
    for i in range(len(parts)-1):
        part = parts[i]
        if len(part) > 2:
            over -= len(part) - 2
            part = part[0] + '-'
            parts[i] = part
            if over <= 0:
                break
    return '.'.join(parts)


def configure_logger(level=None):
    global LOGGING_LEVEL
    if level is None:
        level = LOGGING_LEVEL
    else:
        LOGGING_LEVEL = level
    logger.configure(handlers=[dict(sink=sys.stderr, format=LOGGING_FORMAT, level=level)],
                     patcher=lambda record: record['extra'].update(shortname=shorten_name(record['name'])))
    LoguruInterceptHandler.install()
    return logger


def name_patch(logger, name):
    return logger.patch(lambda record, name=name: (record.update(name=name),
                                                   record['extra'].update(shortname=shorten_name(name))))



"""
Flitter built-in filter shaders
"""

from . import ProgramNode
from .glsl import TemplateLoader
from ...model import null, Matrix33


class Shader(ProgramNode):
    pass


class Transform(Shader):
    DEFAULT_FRAGMENT_SOURCE = TemplateLoader.get_template('transform.frag')

    def render(self, node, references, **kwargs):
        transform_matrix = Matrix33.identity()
        for name, value in node.items():
            if name == 'scale' and (matrix := Matrix33.scale(value)) is not None:
                transform_matrix @= matrix
            elif name == 'translate' and (matrix := Matrix33.translate(value)) is not None:
                transform_matrix @= matrix
            elif name == 'rotate' and (matrix := Matrix33.rotate(value)) is not None:
                transform_matrix @= matrix
            elif name == 'keystone':
                key_x, key_y = value.match(2, float, (0, 0))
                matrix = Matrix33([1, 0, key_x/self.width, 0, 1, key_y/self.height, 0, 0, 1])
                transform_matrix @= matrix
        super().render(node, references, transform_matrix=transform_matrix.inverse(), **kwargs)


class Vignette(Shader):
    DEFAULT_FRAGMENT_SOURCE = TemplateLoader.get_template('vignette.frag')
    EASE_FUNCTIONS = {'linear', 'quad', 'cubic'}

    def render(self, node, references, **kwargs):
        ease = node.get('fade', 1, str)
        if ease not in self.EASE_FUNCTIONS:
            ease = 'linear'
        super().render(node, references, inset=0.25, ease=ease, **kwargs)


class Adjust(Shader):
    DEFAULT_FRAGMENT_SOURCE = TemplateLoader.get_template('adjust.frag')
    TONEMAP_FUNCTIONS = {'reinhard', 'aces'}

    def render(self, node, references, **kwargs):
        tonemap = node.get('tonemap', 1, str)
        if tonemap not in self.TONEMAP_FUNCTIONS:
            tonemap = None
        super().render(node, references, exposure=0, contrast=1, brightness=0, shadows=0, highlights=0,
                       hue=0, saturation=1, color_matrix=(1, 0, 0, 0, 1, 0, 0, 0, 1),
                       gamma=1, tonemap_function=tonemap, **kwargs)


class Blur(Shader):
    DEFAULT_FRAGMENT_SOURCE = TemplateLoader.get_template('blur.frag')

    def render(self, node, references, **kwargs):
        child_textures = self.child_textures
        if len(child_textures) > 1:
            passes = 3
        elif child_textures:
            passes = 2
        else:
            passes = 1
        super().render(node, references, passes=passes, radius=0, sigma=0.3, repeat=(False, False), **kwargs)


class Bloom(Shader):
    DEFAULT_FRAGMENT_SOURCE = TemplateLoader.get_template('bloom.frag')

    def render(self, node, references, **kwargs):
        child_textures = self.child_textures
        if len(child_textures) > 1:
            passes = 5
            downsample_passes = {1, 2, 3}
        elif child_textures:
            passes = 4
            downsample_passes = {0, 1, 2}
        else:
            passes = 1
            downsample_passes = ()
        super().render(node, references, passes=passes, downsample_passes=downsample_passes,
                       radius=0, sigma=0.3, exposure=-1, contrast=1, brightness=0, shadows=0, highlights=0,
                       repeat=(False, False), **kwargs)


class Edges(Shader):
    DEFAULT_FRAGMENT_SOURCE = TemplateLoader.get_template('edges.frag')

    def render(self, node, references, **kwargs):
        child_textures = self.child_textures
        if len(child_textures) > 1:
            passes = 4
            downsample_passes = {1, 2}
        elif child_textures:
            passes = 3
            downsample_passes = {0, 1}
        else:
            passes = 1
            downsample_passes = ()
        super().render(node, references, passes=passes, downsample_passes=downsample_passes,
                       radius=0, sigma=0.3, mixer=0.0, repeat=(False, False), **kwargs)


class Feedback(Shader):
    DEFAULT_FRAGMENT_SOURCE = TemplateLoader.get_template('feedback.frag')

    def render(self, node, references, delta=0, **kwargs):
        if self._target is None:
            delta = 0
        super().render(node, references, delta=delta, mixer=0, timebase=1, glow=0, translate=0, scale=1, rotate=0, repeat=(False, False), **kwargs)


class Flare(Shader):
    DEFAULT_FRAGMENT_SOURCE = TemplateLoader.get_template('flare.frag')

    def render(self, node, references, **kwargs):
        child_textures = self.child_textures
        if len(child_textures) > 1:
            passes = 5
            downsample_passes = {1, 2, 3}
        elif child_textures:
            passes = 4
            downsample_passes = {0, 1, 2}
        else:
            passes = 1
            downsample_passes = ()
        super().render(node, references, passes=passes, downsample_passes=downsample_passes,
                       upright_length=0.25, diagonal_length=0.125, ghosts=6, threshold=1, attenuation=2, aberration=1, **kwargs)


class Noise(Shader):
    DEFAULT_FRAGMENT_SOURCE = TemplateLoader.get_template('noise.frag')

    def render(self, node, references, **kwargs):
        seed_hash = hash(node['seed'] if 'seed' in node else null) / (1 << 48)
        default_values = node.get('default', 4, float, (1, 1, 1, 1))
        super().render(node, references, seed_hash=seed_hash, components=1, octaves=1, roughness=0.5, origin=0, z=0,
                       scale=1, tscale=1, multiplier=0.5, offset=0.5, default_values=default_values, **kwargs)



"""
Tests of the flitter language built-in functions
"""

import math
import struct
import unittest.mock
import zlib

import numpy as np

from flitter.model import Vector, Quaternion, null, Context
from flitter.language.functions import (uniform, normal, beta,
                                        lenv, sumv, accumulate, minv, maxv, minindex, maxindex, mapv, clamp, zipv, count,
                                        roundv, absv, expv, sqrtv, logv, log2v, log10v, ceilv, floorv, fract,
                                        cosv, acosv, sinv, asinv, tanv, hypot, normalize, polar, angle, length, cross, dot,
                                        quaternion, qmul, qbetween, slerp,
                                        split, ordv, chrv,
                                        colortemp, oklab, oklch,
                                        sample, glob, read_text, read_csv, read_bytes)
from flitter.language.noise import noise, octnoise

from . import utils


Tau = 2*math.pi


class TestUniform(utils.TestCase):
    FACTORY = uniform
    DISTRIBUTION = ('uniform',)
    LOWER = 0
    UPPER = 1
    ALPHA = 0.1

    def test_creation(self):
        self.assertIsInstance(self.FACTORY(), self.FACTORY)
        self.assertEqual(hash(self.FACTORY()), hash(Vector()))
        self.assertEqual(hash(self.FACTORY(1.0)), hash(self.FACTORY(1.9)))

    def test_repr(self):
        source = self.FACTORY(1)
        self.assertEqual(repr(source), f"{self.FACTORY.__name__}({hash(source)!r})")

    def test_eq(self):
        source1 = self.FACTORY(1)
        source2 = self.FACTORY(1)
        self.assertIsNot(source1, source2)
        self.assertEqual(source1, source2)

    def test_null(self):
        self.assertEqual(self.FACTORY() + 1, null)
        self.assertEqual(1 + self.FACTORY(), null)
        self.assertEqual(self.FACTORY() * 1, null)
        self.assertEqual(self.FACTORY() / 1, null)
        self.assertEqual(self.FACTORY() % 1, null)
        self.assertEqual(self.FACTORY() // 1, null)
        self.assertEqual(self.FACTORY() ** 1, null)

    def test_bool(self):
        self.assertEqual(bool(self.FACTORY()), True)
        self.assertEqual(bool(self.FACTORY(1)), True)
        self.assertEqual(bool(self.FACTORY("hello")), True)

    def test_null_slice(self):
        source = self.FACTORY()
        self.assertEqual(source[null], null)

    def test_indexing_and_overlap(self):
        source = self.FACTORY()
        last_xs = None
        for i in range(-100, 100):
            source_i = source[i]
            self.assertIsInstance(source_i, float)
            xs = source[i:i+10]
            self.assertIsInstance(xs, Vector)
            self.assertEqual(len(xs), 10)
            self.assertEqual(xs[0], source_i)
            for x in xs:
                self.assertIsInstance(x, float)
            if last_xs is not None:
                self.assertNotEqual(last_xs, xs)
                self.assertEqual(last_xs[1:10], xs[0:9])
            last_xs = xs

    def test_reproducability(self):
        source1 = self.FACTORY(1.0)
        source2 = self.FACTORY(1.1)
        self.assertIsNot(source1, source2)
        self.assertEqual(source1[:10_000], source2[:10_000])

    def test_distribution(self):
        from scipy.stats import kstest
        for i in range(5):
            seed = Vector.symbol('seed').concat(Vector(i))
            with self.subTest(seed=seed):
                result = kstest(self.FACTORY(seed)[:100_000], *self.DISTRIBUTION)
                self.assertGreater(result.pvalue, self.ALPHA)

    def test_range(self):
        source = self.FACTORY()
        for i in range(10000):
            x = source[i]
            self.assertGreater(x, self.LOWER)
            self.assertLess(x, self.UPPER)

    def test_apparent_entropy(self):
        """Should get an average of something like 59 bits of pseudo-entropy per 64"""
        data = struct.pack('<10000Q', *(int((n - self.LOWER) / (self.UPPER - self.LOWER) * (1 << 64)) for n in self.FACTORY()[:10000]))
        compressed = zlib.compress(data, 9)
        self.assertGreater(len(compressed) / len(data), 59/64)


class TestBeta(TestUniform):
    FACTORY = beta
    DISTRIBUTION = ('beta', (2, 2))


class TestNormal(TestUniform):
    FACTORY = normal
    DISTRIBUTION = ('norm',)
    UPPER = math.sqrt(-2 * math.log(1 / (1 << 64)))
    LOWER = -UPPER


class TestNoise(utils.TestCase):
    def test_zero_behaviour(self):
        self.assertEqual(noise(Vector.symbol('seed'), Vector(0)), Vector(0))
        self.assertEqual(noise(Vector.symbol('seed'), Vector(0), Vector(0)), Vector(0))
        n3 = abs(float(noise(Vector.symbol('seed'), Vector(0), Vector(0), Vector(0))))
        self.assertGreater(abs(n3), Vector(0))
        self.assertLess(abs(n3), Vector(1e-30))

    def test_null_behaviour(self):
        self.assertEqual(noise(null, Vector(0)), Vector(0))
        self.assertEqual(noise(Vector.symbol('seed'), null), null)
        self.assertEqual(noise(Vector.symbol('seed'), Vector(0), null), null)
        self.assertEqual(noise(Vector.symbol('seed'), Vector(0), Vector(0), null), null)
        self.assertEqual(octnoise(null, Vector(1), Vector(0.5), Vector(0)), Vector(0))
        self.assertEqual(octnoise(Vector.symbol('seed'), null, Vector(0.5), Vector(0)), null)
        self.assertEqual(octnoise(Vector.symbol('seed'), Vector(1), null, Vector(0)), null)
        self.assertEqual(octnoise(Vector.symbol('seed'), Vector(1), Vector(0.5), null), null)
        self.assertEqual(octnoise(Vector.symbol('seed'), Vector(1), Vector(0.5), Vector(0), null), null)
        self.assertEqual(octnoise(Vector.symbol('seed'), Vector(1), Vector(0.5), Vector(0), Vector(0), null), null)

    def test_reproducability(self):
        self.assertAlmostEqual(noise(null, Vector(0.49)), 0.498956342)
        self.assertAlmostEqual(noise(null, Vector(0.49), Vector(0.49)), 0.167789989)
        self.assertAlmostEqual(noise(null, Vector(0.49), Vector(0.49), Vector(0.49)), -0.219814457)
        self.assertAlmostEqual(noise(null, Vector(0.49), Vector(0.49), Vector(0.49), Vector(0.49)), 0.119003482)

    def test_noise_1(self):
        seed1 = Vector.symbol('seed').concat(Vector(1))
        seed2 = Vector.symbol('seed').concat(Vector(2))
        last_n1 = None
        for x in map(lambda x: x/49.999, range(1, 1001)):
            n1 = float(noise(seed1, Vector(x)))
            self.assertTrue(-1 <= n1 <= 1, f"Noise value bounds failure: {n1}")
            n2 = float(noise(seed2, Vector(x)))
            self.assertTrue(-1 <= n2 <= 1, f"Noise value bounds failure: {n2}")
            self.assertNotEqual(n1, n2)
            if last_n1 is not None:
                self.assertLess(abs(n1 - last_n1), 0.05)
            last_n1 = n1

    def test_noise_2(self):
        seed1 = Vector.symbol('seed').concat(Vector(1))
        seed2 = Vector.symbol('seed').concat(Vector(2))
        last_n1 = None
        y = 1/49.999
        for x in map(lambda x: x/49.999, range(1, 1001)):
            n1 = float(noise(seed1, Vector(x), Vector(y)))
            self.assertTrue(-1 <= n1 <= 1, f"Noise value bounds failure: {n1}")
            n2 = float(noise(seed2, Vector(x), Vector(y)))
            self.assertTrue(-1 <= n2 <= 1, f"Noise value bounds failure: {n2}")
            self.assertNotEqual(n1, n2)
            if last_n1 is not None:
                self.assertLess(abs(n1 - last_n1), 0.05)
            last_n1 = n1
        for y in map(lambda y: y/49.999, range(1, 1001)):
            n1 = float(noise(seed1, Vector(x), Vector(y)))
            self.assertTrue(-1 <= n1 <= 1, f"Noise value bounds failure: {n1}")
            n2 = float(noise(seed2, Vector(x), Vector(y)))
            self.assertTrue(-1 <= n2 <= 1, f"Noise value bounds failure: {n2}")
            self.assertNotEqual(n1, n2)
            self.assertLess(abs(n1 - last_n1), 0.05)
            last_n1 = n1

    def test_noise_3(self):
        seed1 = Vector.symbol('seed').concat(Vector(1))
        seed2 = Vector.symbol('seed').concat(Vector(2))
        last_n1 = None
        y = 1/49.999
        z = 1/49.999
        for x in map(lambda x: x/49.999, range(1, 1001)):
            n1 = float(noise(seed1, Vector(x), Vector(y), Vector(z)))
            self.assertTrue(-1 <= n1 <= 1, f"Noise value bounds failure: {n1}")
            n2 = float(noise(seed2, Vector(x), Vector(y), Vector(z)))
            self.assertTrue(-1 <= n2 <= 1, f"Noise value bounds failure: {n2}")
            self.assertNotEqual(n1, n2)
            if last_n1 is not None:
                self.assertLess(abs(n1 - last_n1), 0.05)
            last_n1 = n1
        for y in map(lambda y: y/49.999, range(1, 1001)):
            n1 = float(noise(seed1, Vector(x), Vector(y), Vector(z)))
            self.assertTrue(-1 <= n1 <= 1, f"Noise value bounds failure: {n1}")
            n2 = float(noise(seed2, Vector(x), Vector(y), Vector(z)))
            self.assertTrue(-1 <= n2 <= 1, f"Noise value bounds failure: {n2}")
            self.assertNotEqual(n1, n2)
            self.assertLess(abs(n1 - last_n1), 0.05)
            last_n1 = n1
        for z in map(lambda z: z/49.999, range(1, 1001)):
            n1 = float(noise(seed1, Vector(x), Vector(y), Vector(z)))
            self.assertTrue(-1 <= n1 <= 1, f"Noise value bounds failure: {n1}")
            n2 = float(noise(seed2, Vector(x), Vector(y), Vector(z)))
            self.assertTrue(-1 <= n2 <= 1, f"Noise value bounds failure: {n2}")
            self.assertNotEqual(n1, n2)
            self.assertLess(abs(n1 - last_n1), 0.05)
            last_n1 = n1

    def test_noise_4(self):
        seed1 = Vector.symbol('seed').concat(Vector(1))
        seed2 = Vector.symbol('seed').concat(Vector(2))
        last_n1 = None
        y = 1/49.999
        z = 1/49.999
        w = 1/49.999
        for x in map(lambda x: x/49.999, range(1, 1001)):
            n1 = float(noise(seed1, Vector(x), Vector(y), Vector(z), Vector(w)))
            self.assertTrue(-1 <= n1 <= 1, f"Noise value bounds failure: {n1}")
            n2 = float(noise(seed2, Vector(x), Vector(y), Vector(z), Vector(w)))
            self.assertTrue(-1 <= n2 <= 1, f"Noise value bounds failure: {n2}")
            self.assertNotEqual(n1, n2)
            if last_n1 is not None:
                self.assertLess(abs(n1 - last_n1), 0.05)
            last_n1 = n1
        for y in map(lambda y: y/49.999, range(1, 1001)):
            n1 = float(noise(seed1, Vector(x), Vector(y), Vector(z), Vector(w)))
            self.assertTrue(-1 <= n1 <= 1, f"Noise value bounds failure: {n1}")
            n2 = float(noise(seed2, Vector(x), Vector(y), Vector(z), Vector(w)))
            self.assertTrue(-1 <= n2 <= 1, f"Noise value bounds failure: {n2}")
            self.assertNotEqual(n1, n2)
            self.assertLess(abs(n1 - last_n1), 0.05)
            last_n1 = n1
        for z in map(lambda z: z/49.999, range(1, 1001)):
            n1 = float(noise(seed1, Vector(x), Vector(y), Vector(z), Vector(w)))
            self.assertTrue(-1 <= n1 <= 1, f"Noise value bounds failure: {n1}")
            n2 = float(noise(seed2, Vector(x), Vector(y), Vector(z), Vector(w)))
            self.assertTrue(-1 <= n2 <= 1, f"Noise value bounds failure: {n2}")
            self.assertNotEqual(n1, n2)
            self.assertLess(abs(n1 - last_n1), 0.05)
            last_n1 = n1
        for w in map(lambda w: w/49.999, range(1, 1001)):
            n1 = float(noise(seed1, Vector(x), Vector(y), Vector(z), Vector(w)))
            self.assertTrue(-1 <= n1 <= 1, f"Noise value bounds failure: {n1}")
            n2 = float(noise(seed2, Vector(x), Vector(y), Vector(z), Vector(w)))
            self.assertTrue(-1 <= n2 <= 1, f"Noise value bounds failure: {n2}")
            self.assertNotEqual(n1, n2)
            self.assertLess(abs(n1 - last_n1), 0.05)
            last_n1 = n1

    def test_octnoise_3(self):
        seed1 = Vector.symbol('seed').concat(Vector(1))
        seed2 = Vector.symbol('seed').concat(Vector(2))
        last_n1 = None
        y = 1/49.999
        z = 1/49.999
        octaves = Vector(3)
        roughness = Vector(0.5)
        for x in map(lambda x: x/49.999, range(1, 1001)):
            n1 = float(octnoise(seed1, octaves, roughness, Vector(x), Vector(y), Vector(z)))
            self.assertTrue(-1 <= n1 <= 1)
            n2 = float(octnoise(seed2, octaves, roughness, Vector(x), Vector(y), Vector(z)))
            self.assertTrue(-1 <= n2 <= 1)
            self.assertNotEqual(n1, n2)
            if last_n1 is not None:
                self.assertLess(abs(n1 - last_n1), 0.1)
            last_n1 = n1
        for y in map(lambda y: y/49.999, range(1, 1001)):
            n1 = float(octnoise(seed1, octaves, roughness, Vector(x), Vector(y), Vector(z)))
            self.assertTrue(-1 <= n1 <= 1)
            n2 = float(octnoise(seed2, octaves, roughness, Vector(x), Vector(y), Vector(z)))
            self.assertTrue(-1 <= n2 <= 1)
            self.assertNotEqual(n1, n2)
            self.assertLess(abs(n1 - last_n1), 0.1)
            last_n1 = n1
        for z in map(lambda z: z/49.999, range(1, 1001)):
            n1 = float(octnoise(seed1, octaves, roughness, Vector(x), Vector(y), Vector(z)))
            self.assertTrue(-1 <= n1 <= 1)
            n2 = float(octnoise(seed2, octaves, roughness, Vector(x), Vector(y), Vector(z)))
            self.assertTrue(-1 <= n2 <= 1)
            self.assertNotEqual(n1, n2)
            self.assertLess(abs(n1 - last_n1), 0.1)
            last_n1 = n1

    def test_seeds(self, n=1001):
        values = sorted([float(noise(Vector(seed), Vector(50/49.999))) for seed in range(n)])
        self.assertLess(values[0], -0.5)
        self.assertTrue(-0.5 < values[n//4] < -0.25)
        self.assertLess(abs(values[n // 2]), 0.05)
        self.assertTrue(0.25 < values[-n//4] < 0.5)
        self.assertGreater(values[-1], 0.5)

    def test_single_range(self):
        values1 = noise(Vector.symbol('seed'), Vector.range(100))
        values2 = [float(noise(Vector.symbol('seed'), Vector(x))) for x in range(100)]
        self.assertEqual(values1, values2)

    def test_double_range(self):
        values1 = noise(Vector.symbol('seed'), Vector.range(10), Vector.range(10))
        values2 = [float(noise(Vector.symbol('seed'), Vector(x), Vector(y))) for x in range(10) for y in range(10)]
        self.assertEqual(values1, values2)

    def test_triple_range(self):
        values1 = noise(Vector.symbol('seed'), Vector.range(10), Vector.range(10), Vector.range(10))
        values2 = [float(noise(Vector.symbol('seed'), Vector(x), Vector(y), Vector(z))) for x in range(10) for y in range(10) for z in range(10)]
        self.assertEqual(values1, values2)


class TestBasicVectorFunctions(utils.TestCase):
    def test_len(self):
        self.assertEqual(lenv(null), 0)
        self.assertEqual(lenv(Vector(1)), 1)
        self.assertEqual(lenv(Vector('hello')), 1)
        self.assertEqual(lenv(Vector(['hello', 'world'])), 2)
        self.assertEqual(lenv(Vector.range(1000)), 1000)

    def test_sum(self):
        xs = Vector.range(10)
        self.assertEqual(sumv(null), null)
        self.assertEqual(sumv(Vector('hello')), null)
        self.assertEqual(sumv(xs, null), null)
        self.assertEqual(sumv(xs, Vector('hello')), null)
        self.assertEqual(sumv(xs, xs), null)
        self.assertEqual(sumv(xs, Vector(0)), null)
        self.assertEqual(sumv(xs), Vector(45))
        self.assertEqual(sumv(xs, Vector(2)), Vector([20, 25]))

    def test_accumulate(self):
        xs = Vector.range(10)
        self.assertEqual(accumulate(null), null)
        self.assertEqual(accumulate(Vector('hello')), null)
        self.assertEqual(accumulate(xs, null), null)
        self.assertEqual(accumulate(xs, Vector('hello')), null)
        self.assertEqual(accumulate(xs, xs), null)
        self.assertEqual(accumulate(xs, Vector(0)), null)
        self.assertEqual(accumulate(xs), Vector([0, 1, 3, 6, 10, 15, 21, 28, 36, 45]))
        self.assertEqual(accumulate(xs, Vector(2)), Vector([0, 1, 2, 4, 6, 9, 12, 16, 20, 25]))

    def test_min(self):
        xs = Vector([10, 4, 9.5, -3, -3.01, 10.01, 3])
        self.assertEqual(minv(null), null)
        self.assertEqual(minv(xs, null), null)
        self.assertEqual(minv(xs, Vector('hello')), null)
        self.assertEqual(minv(xs), -3.01)
        self.assertEqual(minv(*[Vector(x) for x in xs]), -3.01)
        xs = Vector(['b', 'c', 'aa', 'a', 'z', 'zz', 'x', 'w'])
        self.assertEqual(minv(xs), 'a')
        self.assertEqual(minv(*[Vector(x) for x in xs]), 'a')

    def test_minindex(self):
        xs = Vector([10, 4, 9.5, -3, -3.01, 10.01, 3])
        self.assertEqual(minindex(null), null)
        self.assertEqual(minindex(xs, null), 1)
        self.assertEqual(minindex(xs, Vector('hello')), null)
        self.assertEqual(minindex(xs), 4)
        self.assertEqual(minindex(*[Vector(x) for x in xs]), 4)
        xs = Vector(['b', 'c', 'aa', 'a', 'z', 'zz', 'x', 'w'])
        self.assertEqual(minindex(xs), 3)
        self.assertEqual(minindex(*[Vector(x) for x in xs]), 3)

    def test_max(self):
        xs = Vector([10, 4, 9.5, -3, -3.01, 10.01, 3])
        self.assertEqual(maxv(null), null)
        self.assertEqual(maxv(xs, null), xs)
        self.assertEqual(maxv(xs, Vector('hello')), null)
        self.assertEqual(maxv(xs), 10.01)
        self.assertEqual(maxv(*[Vector(x) for x in xs]), 10.01)
        xs = Vector(['b', 'c', 'aa', 'a', 'z', 'zz', 'x', 'w'])
        self.assertEqual(maxv(xs), 'zz')
        self.assertEqual(maxv(*[Vector(x) for x in xs]), 'zz')

    def test_maxindex(self):
        xs = Vector([10, 4, 9.5, -3, -3.01, 10.01, 3])
        self.assertEqual(maxindex(null), null)
        self.assertEqual(maxindex(xs, null), 0)
        self.assertEqual(maxindex(xs, Vector('hello')), null)
        self.assertEqual(maxindex(xs), 5)
        self.assertEqual(maxindex(*[Vector(x) for x in xs]), 5)
        xs = Vector(['b', 'c', 'aa', 'a', 'z', 'zz', 'x', 'w'])
        self.assertEqual(maxindex(xs), 5)
        self.assertEqual(maxindex(*[Vector(x) for x in xs]), 5)

    def test_map(self):
        self.assertEqual(mapv(null, Vector(-1), Vector(1)), null)
        self.assertEqual(mapv(Vector(0.5), null, Vector(1)), null)
        self.assertEqual(mapv(Vector(0.5), Vector(-1), null), null)
        self.assertEqual(mapv(Vector(0.5), Vector(-1), Vector(1)), Vector(0))
        self.assertEqual(mapv(Vector(0.5), Vector([-1, 0]), Vector(1)), Vector([0, 0.5]))
        self.assertEqual(mapv(Vector(0.5), Vector([-1, 0]), Vector([1, 2])), Vector([0, 1]))

    def test_clamp(self):
        self.assertEqual(clamp(null, Vector(-1), Vector(1)), null)
        self.assertEqual(clamp(Vector(0.5), null, Vector(1)), null)
        self.assertEqual(clamp(Vector(0.5), Vector(-1), null), null)
        self.assertEqual(clamp(Vector(0.5), Vector(-1), Vector(1)), Vector(0.5))
        self.assertEqual(clamp(Vector(-2), Vector(-1), Vector(1)), Vector(-1))
        self.assertEqual(clamp(Vector(2), Vector(-1), Vector(1)), Vector(1))
        self.assertEqual(clamp(Vector([-2, 0.5, 2]), Vector(-1), Vector(1)), Vector([-1, 0.5, 1]))
        self.assertEqual(clamp(Vector([-2, 0.5, 2]), Vector([-1, 1]), Vector([1, 2])), Vector([-1, 1, 1]))

    def test_zip(self):
        xs = Vector.range(10)
        ys = Vector.range(0.5, 10)
        self.assertEqual(zipv(), null)
        self.assertEqual(zipv(null), null)
        self.assertEqual(zipv(xs), xs)
        self.assertEqual(zipv(xs, ys), Vector.range(0, 10, 0.5))
        self.assertEqual(zipv(*(Vector(x) for x in xs)), xs)
        self.assertEqual(zipv(Vector(['a', 'c', 'e']), Vector(['b', 'd'])), Vector(['a', 'b', 'c', 'd', 'e', 'b']))
        self.assertEqual(zipv(Vector(['a', 'c', 'e']), Vector([0, 1])), Vector(['a', 0, 'c', 1, 'e', 0]))

    def test_count(self):
        xs = Vector.range(10)
        self.assertEqual(count(null, null), null)
        self.assertEqual(count(xs, null), [0]*10)
        self.assertEqual(count(null, xs), null)
        self.assertEqual(count(Vector(0), xs), 1)
        self.assertEqual(count(Vector(10), xs), 0)
        self.assertEqual(count(xs, xs), [1]*10)
        self.assertEqual(count(xs, xs.concat(xs)), [2]*10)
        self.assertEqual(count(Vector(['a', 1.0]), xs), [0, 1])
        self.assertEqual(count(Vector(1), Vector(['a', 'b', 1, 2, 3, 1])), 2)
        self.assertEqual(count(Vector(['a', 1]), Vector(['a', 'b', 1, 2, 3, 1])), [1, 2])


class TestUnaryMathFunctions(utils.TestCase):
    def assertMatchesUnaryFunc(self, vfunc, func, xs=None, places=None, delta=None):
        self.assertEqual(vfunc(null), null)
        self.assertEqual(vfunc(Vector('hello')), null)
        xs = Vector.range(-10, 10, 0.2) if xs is None else xs
        ys = [func(x) for x in xs]
        self.assertAllAlmostEqual([float(vfunc(xs.item(i))) for i in range(len(xs))], ys, places=places, delta=delta)
        self.assertAllAlmostEqual(vfunc(xs), ys, places=places, delta=delta)

    def test_round(self):
        self.assertMatchesUnaryFunc(roundv, round)

    def test_abs(self):
        self.assertMatchesUnaryFunc(absv, abs)

    def test_exp(self):
        self.assertMatchesUnaryFunc(expv, math.exp)

    def test_sqrt(self):
        self.assertMatchesUnaryFunc(sqrtv, math.sqrt, Vector.range(0, 10, 0.1))

    def test_log(self):
        self.assertMatchesUnaryFunc(logv, math.log, Vector.range(0.001, 10.001, 0.1))

    def test_log2(self):
        self.assertMatchesUnaryFunc(log2v, math.log2, Vector.range(0.001, 10.001, 0.1))

    def test_log10(self):
        self.assertMatchesUnaryFunc(log10v, math.log10, Vector.range(0.001, 10.001, 0.1))

    def test_ceil(self):
        self.assertMatchesUnaryFunc(ceilv, math.ceil)

    def test_floor(self):
        self.assertMatchesUnaryFunc(floorv, math.floor)

    def test_fract(self):
        self.assertMatchesUnaryFunc(fract, lambda x: x % 1)


class TestTrig(utils.TestCase):
    def setUp(self):
        self.a = Vector([1, 2, 3, 4])
        self.b = Vector([4, 5, 6, 7])
        self.c = Vector([2])

    def test_cos(self):
        self.assertEqual(cosv(null), null)
        self.assertEqual(cosv(Vector('hello')), null)
        theta = Vector.range(0, 1, 0.01)
        values = [math.cos(th) for th in theta*Tau]
        for i in range(len(values)):
            self.assertEqual(cosv(theta.item(i)), values[i])
        self.assertEqual(cosv(theta), values)

    def test_acos(self):
        self.assertEqual(acosv(null), null)
        self.assertEqual(acosv(Vector('hello')), null)
        xs = Vector.range(-1, 1, 0.02)
        values = [math.acos(x)/Tau for x in xs]
        for i in range(len(values)):
            self.assertEqual(acosv(xs.item(i)), values[i])
        self.assertAllAlmostEqual(acosv(xs), values)

    def test_sin(self):
        self.assertEqual(sinv(null), null)
        self.assertEqual(sinv(Vector('hello')), null)
        theta = Vector.range(0, 1, 0.01)
        values = [math.sin(th) for th in theta*Tau]
        for i in range(len(values)):
            self.assertEqual(sinv(theta.item(i)), values[i])
        self.assertEqual(sinv(theta), values)

    def test_asin(self):
        self.assertEqual(asinv(null), null)
        self.assertEqual(asinv(Vector('hello')), null)
        xs = Vector.range(-1, 1, 0.02)
        values = [math.asin(x)/Tau for x in xs]
        for i in range(len(values)):
            self.assertEqual(asinv(xs.item(i)), values[i])
        self.assertAllAlmostEqual(asinv(xs), values)

    def test_tan(self):
        self.assertEqual(tanv(null), null)
        self.assertEqual(tanv(Vector('hello')), null)
        theta = Vector.range(0, 1, 0.01)
        values = [math.tan(th) for th in theta*Tau]
        for i in range(len(values)):
            self.assertEqual(tanv(theta.item(i)), values[i])
        self.assertEqual(tanv(theta), values)

    def test_hypot(self):
        self.assertEqual(hypot(), null)
        self.assertEqual(hypot(null), null)
        self.assertEqual(hypot(self.a, null), null)
        self.assertEqual(hypot(Vector([3, 4])), Vector(5))
        self.assertEqual(hypot(Vector([3, 4, 5])), Vector(math.sqrt(50)))
        self.assertAllAlmostEqual(hypot(self.a, self.b), Vector([math.sqrt(17), math.sqrt(29), math.sqrt(45), math.sqrt(65)]))
        self.assertAllAlmostEqual(hypot(self.a, self.c), Vector([math.sqrt(5), math.sqrt(8), math.sqrt(13), math.sqrt(20)]))
        self.assertAllAlmostEqual(hypot(self.a, self.b, self.c),
                                  Vector([math.sqrt(21), math.sqrt(33), math.sqrt(49), math.sqrt(69)]))

    def test_normalize(self):
        self.assertEqual(normalize(null), null)
        self.assertEqual(normalize(Vector([3, 4])), Vector([3, 4]) / Vector(5))
        self.assertEqual(normalize(Vector([3, 4, 5])), Vector([3, 4, 5]) / Vector(math.sqrt(50)))

    def test_cross(self):
        self.assertEqual(cross(null, null), null)
        self.assertEqual(cross(Vector([3, 4]), Vector([3, 4, 5])), null)
        self.assertEqual(cross(Vector([3, 4, 5]), Vector([3, 4])), null)
        self.assertEqual(cross(Vector([3, 4, 5]), Vector([3, 4, 5])), Vector([0, 0, 0]))
        self.assertEqual(cross(Vector([1, 0, 0]), Vector([0, 1, 0])), Vector([0, 0, 1]))
        self.assertEqual(cross(Vector([0, 1, 0]), Vector([0, 0, 1])), Vector([1, 0, 0]))
        self.assertEqual(cross(Vector([0, 0, 1]), Vector([1, 0, 0])), Vector([0, 1, 0]))

    def test_dot(self):
        self.assertEqual(dot(null, null), null)
        self.assertEqual(dot(Vector([3, 4]), Vector([3, 4, 5])), Vector(3*3 + 4*4 + 3*5))
        self.assertEqual(dot(Vector([3, 4, 5]), Vector([3, 4])), Vector(3*3 + 4*4 + 5*3))
        self.assertEqual(dot(Vector([3, 4, 5]), Vector([5, 4, 3])), Vector(3*5 + 4*4 + 5*3))

    def test_polar(self):
        self.assertEqual(polar(null), null)
        self.assertEqual(polar(Vector('hello')), null)
        theta = Vector.range(0, 1, 0.01)
        values = [(math.cos(th), math.sin(th)) for th in theta*Tau]
        for i in range(len(values)):
            self.assertEqual(polar(theta.item(i)), values[i])
        self.assertEqual(polar(theta), Vector.compose(values))

    def test_angle(self):
        self.assertEqual(angle(null), null)
        self.assertAllAlmostEqual(angle(self.a), Vector([math.atan2(2, 1)/Tau, math.atan2(4, 3)/Tau]))
        self.assertEqual(angle(self.a, null), null)
        self.assertEqual(angle(null, self.a), null)
        self.assertAllAlmostEqual(angle(self.a, self.b),
                                  Vector([math.atan2(4, 1)/Tau, math.atan2(5, 2)/Tau, math.atan2(6, 3)/Tau, math.atan2(7, 4)/Tau]))
        self.assertAllAlmostEqual(angle(self.a, self.c),
                                  Vector([math.atan2(2, 1)/Tau, math.atan2(2, 2)/Tau, math.atan2(2, 3)/Tau, math.atan2(2, 4)/Tau]))

    def test_length(self):
        self.assertEqual(length(null), null)
        self.assertAllAlmostEqual(length(self.a), Vector([math.sqrt(5), 5]))
        self.assertEqual(length(self.a, null), null)
        self.assertEqual(length(null, self.a), null)
        self.assertAllAlmostEqual(length(self.a, self.b),
                                  Vector([math.sqrt(17), math.sqrt(29), math.sqrt(45), math.sqrt(65)]))
        self.assertAllAlmostEqual(length(self.a, self.c),
                                  Vector([math.sqrt(5), math.sqrt(8), math.sqrt(13), math.sqrt(20)]))


class TestQuaternions(utils.TestCase):
    def test_quaternion(self):
        self.assertEqual(quaternion(Vector([1, 0, 0]), Vector(0)), [1, 0, 0, 0])
        self.assertEqual(quaternion(Vector([0, 1, 0]), Vector(0)), [1, 0, 0, 0])
        self.assertEqual(quaternion(Vector([0, 0, 1]), Vector(0)), [1, 0, 0, 0])
        c, s = math.cos(math.pi*0.25), math.sin(math.pi*0.25)
        self.assertEqual(quaternion(Vector([1, 0, 0]), Vector(0.25)), [c, s, 0, 0])
        self.assertEqual(quaternion(Vector([0, 1, 0]), Vector(0.25)), [c, 0, s, 0])
        self.assertEqual(quaternion(Vector([0, 0, 1]), Vector(0.25)), [c, 0, 0, s])
        self.assertAllAlmostEqual(quaternion(Vector([1, 0, 0]), Vector(0.5)), [0, 1, 0, 0])
        self.assertAllAlmostEqual(quaternion(Vector([0, 1, 0]), Vector(0.5)), [0, 0, 1, 0])
        self.assertAllAlmostEqual(quaternion(Vector([0, 0, 1]), Vector(0.5)), [0, 0, 0, 1])
        c, s = math.cos(math.pi/3), math.sin(math.pi/3)
        self.assertEqual(quaternion(Vector([1, 0, 0]), Vector(1/3)), [c, s, 0, 0])
        self.assertEqual(quaternion(Vector([0, 1, 0]), Vector(1/3)), [c, 0, s, 0])
        self.assertEqual(quaternion(Vector([0, 0, 1]), Vector(1/3)), [c, 0, 0, s])
        self.assertAllAlmostEqual(quaternion(Vector([1, 0, 0]), Vector(1)), Vector([-1, 0, 0, 0]))
        self.assertAllAlmostEqual(quaternion(Vector([0, 1, 0]), Vector(1)), Vector([-1, 0, 0, 0]))
        self.assertAllAlmostEqual(quaternion(Vector([0, 0, 1]), Vector(1)), Vector([-1, 0, 0, 0]))

    def test_qmul(self):
        qx = quaternion(Vector([1, 0, 0]), Vector(0.125))
        qy = quaternion(Vector([0, 1, 0]), Vector(0.125))
        qz = quaternion(Vector([0, 0, 1]), Vector(0.125))
        self.assertAllAlmostEqual(qmul(qx, qx), quaternion(Vector([1, 0, 0]), Vector(0.25)))
        self.assertAllAlmostEqual(qmul(qy, qy), quaternion(Vector([0, 1, 0]), Vector(0.25)))
        self.assertAllAlmostEqual(qmul(qz, qz), quaternion(Vector([0, 0, 1]), Vector(0.25)))
        self.assertAllAlmostEqual(qmul(qx, quaternion(Vector([1, 0, 0]), Vector(-0.125))), [1, 0, 0, 0])
        self.assertAllAlmostEqual(qmul(qy, quaternion(Vector([0, 1, 0]), Vector(-0.125))), [1, 0, 0, 0])
        self.assertAllAlmostEqual(qmul(qz, quaternion(Vector([0, 0, 1]), Vector(-0.125))), [1, 0, 0, 0])
        self.assertNotAllAlmostEqual(qmul(qx, qy), qmul(qy, qx))
        self.assertNotAllAlmostEqual(qmul(qx, qz), qmul(qz, qx))
        self.assertNotAllAlmostEqual(qmul(qy, qz), qmul(qz, qy))
        q = quaternion(Vector([1, 1, 1]), Vector(1/3))
        self.assertAllAlmostEqual(qmul(qmul(q, q), q), [-1, 0, 0, 0])
        self.assertAllAlmostEqual(qmul(qmul(qmul(q, q), q), qmul(qmul(q, q), q)), [1, 0, 0, 0])

    def test_qbetween(self):
        self.assertAllAlmostEqual(qbetween(Vector([1, 0, 0]), Vector([0, 1, 0])), Quaternion.euler([0, 0, 1], 0.25))
        self.assertAllAlmostEqual(qbetween(Vector([0, 0, 1]), Vector([1, 0, 0])), Quaternion.euler([0, 1, 0], 0.25))
        self.assertAllAlmostEqual(qbetween(Vector([0, 1, 0]), Vector([0, 0, 1])), Quaternion.euler([1, 0, 0], 0.25))
        self.assertAllAlmostEqual(qbetween(Vector([1, 0, 0]), Vector([1, 1, 0])), Quaternion.euler([0, 0, 1], 0.125))
        self.assertAllAlmostEqual(qbetween(Vector([0, 0, 1]), Vector([1, 0, 1])), Quaternion.euler([0, 1, 0], 0.125))
        self.assertAllAlmostEqual(qbetween(Vector([0, 1, 0]), Vector([0, 1, 1])), Quaternion.euler([1, 0, 0], 0.125))
        self.assertAllAlmostEqual(qbetween(Vector([1, 0, 0]), Vector([-1, 0, 0])), Quaternion.euler([0, 1, 0], 0.5))
        self.assertAllAlmostEqual(qbetween(Vector([0, 1, 0]), Vector([0, -1, 0])), Quaternion.euler([-1, 0, 0], 0.5))
        self.assertAllAlmostEqual(qbetween(Vector([0, 0, 1]), Vector([0, 0, -1])), Quaternion.euler([0, -1, 0], 0.5))

    def test_slerp(self):
        qx = quaternion(Vector([1, 0, 0]), Vector(0.25))
        qy = quaternion(Vector([0, 1, 0]), Vector(0.25))
        self.assertAllAlmostEqual(slerp(Vector(0), qx, qy), qx)
        self.assertAllAlmostEqual(slerp(Vector(1), qx, qy), qy)
        self.assertAllAlmostEqual(slerp(Vector(-1), qx, qx), qx)
        self.assertAllAlmostEqual(slerp(Vector(0), qx, qx), qx)
        self.assertAllAlmostEqual(slerp(Vector(1), qx, qx), qx)
        self.assertAllAlmostEqual(slerp(Vector(2), qx, qx), qx)


class TestStringFuncs(utils.TestCase):
    def test_ord(self):
        self.assertEqual(ordv(null), null)
        self.assertEqual(ordv(Vector('A')), Vector([65]))
        self.assertEqual(ordv(Vector('AB')), Vector([65, 66]))
        self.assertEqual(ordv(Vector(['A', 'B'])), Vector([65, 66]))

    def test_chr(self):
        self.assertEqual(chrv(null), null)
        self.assertEqual(chrv(Vector('A')), null)
        self.assertEqual(chrv(Vector([65])), Vector(['A']))
        self.assertEqual(chrv(Vector([65, 66])), Vector(['AB']))

    def test_split(self):
        self.assertEqual(split(null), null)
        self.assertEqual(split(Vector(['Hello world!'])), Vector(['Hello world!']))
        self.assertEqual(split(Vector(['Hello world!\n'])), Vector(['Hello world!']))
        self.assertEqual(split(Vector(['Hello\nworld!'])), Vector(['Hello', 'world!']))
        self.assertEqual(split(Vector(['Hello\nworld!\n'])), Vector(['Hello', 'world!']))
        self.assertEqual(split(Vector(['Hello\n\nworld!\n'])), Vector(['Hello', '', 'world!']))


class TestColorFuncs(utils.TestCase):
    def test_colortemp(self):
        self.assertAllAlmostEqual(colortemp(Vector(6503.5)), Vector([1, 1, 1]), places=None, delta=0.05)

    def test_oklab(self):
        self.assertAllAlmostEqual(oklab(Vector([0, 0, 0])), Vector([0, 0, 0]), places=2)
        self.assertAllAlmostEqual(oklab(Vector([1, 0, 0])), Vector([1, 1, 1]), places=2)
        self.assertAllAlmostEqual(oklab(Vector([0.5, +0.4, 0])), Vector([0.87, -0.15, 0.1]), places=2)
        self.assertAllAlmostEqual(oklab(Vector([0.5, -0.4, 0])), Vector([-0.33, 0.31, 0.15]), places=2)
        self.assertAllAlmostEqual(oklab(Vector([0.5, 0, +0.4])), Vector([0.47, 0.02, -0.08]), places=2)
        self.assertAllAlmostEqual(oklab(Vector([0.5, 0, -0.4])), Vector([0.05, -0.07, 1.69]), places=2)

    def test_oklch(self):
        self.assertAllAlmostEqual(oklch(Vector([0, 0, 0])), Vector([0, 0, 0]), places=2)
        self.assertAllAlmostEqual(oklch(Vector([1, 0, 0])), Vector([1, 1, 1]), places=2)
        self.assertAllAlmostEqual(oklch(Vector([0.5, 0.4, 0])), Vector([0.87, -0.15, 0.1]), places=2)
        self.assertAllAlmostEqual(oklch(Vector([0.5, 0.4, 1/3])), Vector([0.09, 0.17, -0.09]), places=2)
        self.assertAllAlmostEqual(oklch(Vector([0.5, 0.4, 2/3])), Vector([-0.15, 0.06, 1.42]), places=2)
        self.assertAllAlmostEqual(oklch(Vector([0.5, 0.4, 1])), Vector([0.87, -0.15, 0.1]), places=2)
        self.assertAllAlmostEqual(oklch(Vector([0.5, 0.4, -1])), Vector([0.87, -0.15, 0.1]), places=2)


class TestSampling(utils.TestCase):
    def test_missing_texture(self):
        context = Context(references={})
        color = sample(context, Vector('test'), Vector([0, 0]))
        self.assertEqual(color, null)
        color = sample(context, Vector('test'), Vector([0, 0]), Vector([1, 2, 3, 4]))
        self.assertEqual(color, Vector([1, 2, 3, 4]))

    def test_no_array(self):
        reference = unittest.mock.Mock()
        del reference.array
        context = Context(references={'test': reference})
        color = sample(context, Vector('test'), Vector([0, 0]), Vector([1, 2, 3, 4]))
        self.assertEqual(color, Vector([1, 2, 3, 4]))

    def test_array_is_none(self):
        reference = unittest.mock.Mock()
        reference.array = None
        context = Context(references={'test': reference})
        color = sample(context, Vector('test'), Vector([0, 0]), Vector([1, 2, 3, 4]))
        self.assertEqual(color, Vector([1, 2, 3, 4]))

    def test_array_is_empty(self):
        reference = unittest.mock.Mock()
        reference.array = np.array([], dtype='float32').reshape((0, 0, 3))
        context = Context(references={'test': reference})
        color = sample(context, Vector('test'), Vector([0, 0]), Vector([1, 2, 3, 4]))
        self.assertEqual(color, Vector([1, 2, 3, 4]))

    def test_4x4_array(self):
        reference = unittest.mock.Mock()
        reference.array = np.array([[[0, 0, 0, 1], [1, 0, 0, 1]], [[0, 1, 0, 1], [1, 1, 0, 1]]], dtype='float32')
        context = Context(references={'test': reference})
        self.assertEqual(sample(context, Vector('test'), Vector([0, 0]), Vector([1, 2, 3, 4])), Vector([0, 0, 0, 1]))
        self.assertEqual(sample(context, Vector('test'), Vector([0.5, 0]), Vector([1, 2, 3, 4])), Vector([1, 0, 0, 1]))
        self.assertEqual(sample(context, Vector('test'), Vector([0, 0.5]), Vector([1, 2, 3, 4])), Vector([0, 1, 0, 1]))
        self.assertEqual(sample(context, Vector('test'), Vector([0.5, 0.5]), Vector([1, 2, 3, 4])), Vector([1, 1, 0, 1]))
        self.assertEqual(sample(context, Vector('test'), Vector([0.5, 0]), Vector([1, 2, 3, 4])), Vector([1, 0, 0, 1]))
        self.assertEqual(sample(context, Vector('test'), Vector([-0.5, 0]), Vector([1, 2, 3, 4])), Vector([1, 2, 3, 4]))
        self.assertEqual(sample(context, Vector('test'), Vector([1.0, 0]), Vector([1, 2, 3, 4])), Vector([1, 2, 3, 4]))
        self.assertEqual(sample(context, Vector('test'), Vector([0, -0.5]), Vector([1, 2, 3, 4])), Vector([1, 2, 3, 4]))
        self.assertEqual(sample(context, Vector('test'), Vector([0, 1.0]), Vector([1, 2, 3, 4])), Vector([1, 2, 3, 4]))


class TestFileFunctions(utils.TestCase):
    def test_glob(self):
        path = unittest.mock.Mock()
        context = Context(path=path)
        a = unittest.mock.Mock()
        b = unittest.mock.Mock()
        c = unittest.mock.Mock()
        a.resolve.return_value = 'a'
        b.resolve.return_value = 'b'
        c.resolve.return_value = 'c'
        path.parent.glob.return_value = [c, a, b]
        self.assertEqual(glob(context, Vector('*')), ['a', 'b', 'c'])
        path.parent.glob.assert_called_with('*')

    def test_read_text(self):
        path = 'world'
        context = Context(path=path)
        with unittest.mock.patch('flitter.language.functions.SharedCache') as cache:
            cache_path = unittest.mock.Mock()
            cache_path.read_text.return_value = 'Hello world!\n'
            cache.get_with_root.return_value = cache_path
            self.assertEqual(read_text(context, Vector('hello')), Vector('Hello world!\n'))
            cache.get_with_root.assert_called_with('hello', 'world')
            cache_path.read_text.assert_called_with(encoding='utf8')

    def test_read_bytes(self):
        path = 'world'
        context = Context(path=path)
        with unittest.mock.patch('flitter.language.functions.SharedCache') as cache:
            cache_path = unittest.mock.Mock()
            cache_path.read_bytes.return_value = b'Hello world!\n'
            cache.get_with_root.return_value = cache_path
            self.assertEqual(read_bytes(context, Vector('hello')), Vector(b'Hello world!\n'))
            cache.get_with_root.assert_called_with('hello', 'world')
            cache_path.read_bytes.assert_called_with()

    def test_read_csv(self):
        path = 'world'
        context = Context(path=path)
        with unittest.mock.patch('flitter.language.functions.SharedCache') as cache:
            cache_path = unittest.mock.Mock()
            cache_path.read_csv_vector.return_value = Vector(['Hello', 10])
            cache.get_with_root.return_value = cache_path
            self.assertEqual(read_csv(context, Vector('hello'), Vector(3)), Vector(['Hello', 10]))
            cache.get_with_root.assert_called_with('hello', 'world')
            cache_path.read_csv_vector.assert_called_with(3)



"""
The main Flitter engine
"""

import asyncio
import gc
from pathlib import Path
import pickle
import time

from loguru import logger

from .. import setproctitle
from ..cache import SharedCache
from ..clock import BeatCounter, system_clock
from ..language.vm import log_vm_stats
from ..model import Vector, StateDict, Context, null, numbers_cache_counts, empty_numbers_cache
from ..plugins import get_plugin
from ..render.window.models import Model


class EngineController:
    STATE_SAVE_PERIOD = 1

    def __init__(self, target_fps=60, screen=0, fullscreen=False, vsync=False, state_file=None,
                 reset_on_switch=False, state_simplify_wait=0, realtime=True, defined_names=None, vm_stats=False,
                 run_time=None, offscreen=False, disable_simplifier=False, opengl_es=False):
        self.default_fps = target_fps
        self.target_fps = target_fps
        self.realtime = realtime
        self.screen = screen
        self.fullscreen = fullscreen
        self.vsync = vsync
        self.offscreen = offscreen
        self.opengl_es = opengl_es
        self.reset_on_switch = reset_on_switch
        self.disable_simplifier = disable_simplifier
        self.state_simplify_wait = 0 if self.disable_simplifier else state_simplify_wait / 2
        if defined_names:
            self.defined_names = {key: Vector.coerce(value) for key, value in defined_names.items()}
        else:
            self.defined_names = {}
        self.vm_stats = vm_stats
        self.run_time = run_time
        self.state_file = Path(state_file) if state_file is not None else None
        if self.state_file is not None and self.state_file.exists():
            logger.info("Recover state from state file: {}", self.state_file)
            with open(self.state_file, 'rb') as file:
                self.global_state = pickle.load(file)
        else:
            self.global_state = {}
        self.global_state_dirty = False
        self.state = None
        self.state_timestamp = None
        self.state_generation0 = None
        self.state_generation1 = None
        self.state_generation2 = None
        self.renderers = {}
        self.counter = BeatCounter()
        self.pages = []
        self.switch_page = None
        self.current_page = None
        self.current_path = None
        self._references = {}

    def load_page(self, filename):
        page_number = len(self.pages)
        path = SharedCache.get_with_root(filename, '.')
        self.pages.append((path, self.global_state.setdefault(str(path), StateDict())))
        logger.info("Added code page {}: {}", page_number, path)
        if page_number == 0:
            self.switch_to_page(0)

    def switch_to_page(self, page_number):
        if self.pages is not None and 0 <= page_number < len(self.pages):
            if self.state is not None:
                if self.reset_on_switch:
                    self.state.clear()
                self.global_state_dirty = self.global_state_dirty or self.state.changed
                self.state.clear_changed()
            path, state = self.pages[page_number]
            self.state = state
            self.state_timestamp = system_clock()
            self.state_generation0 = set()
            self.state_generation1 = set()
            self.state_generation2 = set()
            self.current_path = path
            self.current_page = page_number
            SharedCache.set_root(self.current_path)
            logger.info("Switched to page {}: {}", page_number, self.current_path)
            if counter_state := self.state['_counter']:
                tempo, quantum, start = counter_state
                self.counter.reset(tempo, int(quantum), start)
                logger.info("Restore counter at beat {:.1f}, tempo {:.1f}, quantum {}", self.counter.beat, self.counter.tempo, self.counter.quantum)
            else:
                self.counter.reset()
            setproctitle(f'flitter {self.current_path}')

    def has_next_page(self):
        return self.current_page < len(self.pages) - 1

    def next_page(self):
        if self.current_page < len(self.pages) - 1:
            self.switch_page = self.current_page + 1

    def has_previous_page(self):
        return self.current_page > 0

    def previous_page(self):
        if self.current_page > 0:
            self.switch_page = self.current_page - 1

    async def update_renderers(self, root, **kwargs):
        nodes_by_kind = {}
        for node in root.children:
            nodes_by_kind.setdefault(node.kind, []).append(node)
        tasks = []
        updated_references = []
        for kind, nodes in nodes_by_kind.items():
            renderer_class = get_plugin('flitter.render', kind)
            if renderer_class is not None:
                renderers = self.renderers.setdefault(kind, [])
                count = 0
                for node in nodes:
                    if count == len(renderers):
                        renderer = renderer_class(**kwargs)
                        renderers.append(renderer)
                    references = dict(self._references)
                    tasks.append(asyncio.create_task(renderers[count].update(self, node, references=references, **kwargs)))
                    updated_references.append(references)
                    count += 1
                while len(renderers) > count:
                    await renderers.pop().destroy()
        for kind in list(self.renderers):
            if kind not in nodes_by_kind:
                for renderer in self.renderers.pop(kind):
                    await renderer.destroy()
        try:
            await asyncio.gather(*tasks)
        except Exception:
            for task in tasks:
                if not task.done():
                    task.cancel()
            await asyncio.gather(*tasks, return_exceptions=True)
            raise
        for references in updated_references:
            self._references.update(references)

    def handle_pragmas(self, pragmas, timestamp):
        if '_counter' not in self.state:
            tempo = pragmas.get('tempo', null).match(1, float, 120)
            if tempo != self.counter.tempo:
                self.counter.set_tempo(tempo, timestamp)
            quantum = pragmas.get('quantum', null).match(1, float, 4)
            if quantum != self.counter.quantum:
                self.counter.set_quantum(quantum, timestamp)
        self.target_fps = pragmas.get('fps', null).match(1, float, self.default_fps)

    async def run(self):
        try:
            frame_count = 0
            frames = []
            start_time = frame_time = system_clock() if self.realtime else self.counter.start
            last = self.counter.beat_at_time(frame_time)
            save_state_time = system_clock() + self.STATE_SAVE_PERIOD
            execution = render = housekeeping = 0
            slow_frame = False
            performance = 1 if self.realtime else 2
            run_program = current_program = errors = logs = None
            simplify_state_time = system_clock() + self.state_simplify_wait
            static = dict(self.defined_names)
            static.update({'realtime': self.realtime, 'screen': self.screen, 'fullscreen': self.fullscreen,
                           'vsync': self.vsync, 'offscreen': self.offscreen, 'opengl_es': self.opengl_es, 'run_time': self.run_time})
            gc.disable()
            while self.run_time is None or int(round((frame_time - start_time) * self.target_fps)) < int(round(self.run_time * self.target_fps)):
                housekeeping -= system_clock()

                beat = self.counter.beat_at_time(frame_time)
                delta = beat - last
                last = beat
                dynamic = {'beat': beat, 'quantum': self.counter.quantum, 'tempo': self.counter.tempo, 'fps': self.target_fps,
                           'delta': delta, 'time': frame_time, 'frame': frame_count, 'performance': performance, 'slow_frame': slow_frame,
                           'clock': time.time()}
                names = dict(static)
                names.update(dynamic)

                program = self.current_path.read_flitter_program(static=static, dynamic=dynamic, simplify=not self.disable_simplifier)
                if program is not current_program:
                    level = 'SUCCESS' if current_program is None else 'INFO'
                    logger.log(level, "Loaded page {}: {}", self.current_page, self.current_path)
                    run_program = current_program = program
                    self.handle_pragmas(program.pragmas, frame_time)
                    errors = logs = None
                    self.state_generation0 ^= self.state_generation1
                    self.state_generation1 = self.state_generation2
                    self.state_generation2 = set()
                    simplify_state_time = system_clock() + self.state_simplify_wait

                now = system_clock()
                housekeeping += now
                execution -= now

                if run_program is not None:
                    context = Context(names={key: Vector.coerce(value) for key, value in dynamic.items()},
                                      state=self.state, references=self._references)
                    run_program.run(context, record_stats=self.vm_stats)
                else:
                    context = Context()

                new_errors = context.errors.difference(errors) if errors is not None else context.errors
                errors = context.errors
                for error in new_errors:
                    logger.error("Execution error: {}", error)
                new_logs = context.logs.difference(logs) if logs is not None else context.logs
                logs = context.logs
                for log in new_logs:
                    print(log)

                now = system_clock()
                execution += now
                render -= now

                await self.update_renderers(context.root, **names)

                now = system_clock()
                render += now
                housekeeping -= now

                del context
                SharedCache.clean()

                self.state['_counter'] = self.counter.tempo, self.counter.quantum, self.counter.start

                if self.state.changed:
                    if self.state_simplify_wait:
                        changed_keys = self.state.changed_keys - self.state_generation0
                        self.state_generation0 ^= changed_keys
                        self.state_generation0 &= self.state.keys()
                        if changed_keys:
                            generation1to0 = self.state_generation1 & changed_keys
                            changed_keys -= generation1to0
                            self.state_generation1 -= generation1to0
                            generation2to0 = self.state_generation2 & changed_keys
                            if generation2to0:
                                if run_program is not current_program:
                                    run_program = current_program
                                    logger.debug("Undo simplification on state; original program with {} instructions", len(run_program))
                                self.state_generation1 ^= self.state_generation2 - generation2to0
                                self.state_generation2 = set()
                                simplify_state_time = system_clock() + self.state_simplify_wait
                    self.global_state_dirty = True
                    self.state_timestamp = system_clock()
                    self.state.clear_changed()

                if self.state_simplify_wait and system_clock() > simplify_state_time:
                    if current_program is not None and self.state_generation1:
                        if self.state_generation1:
                            self.state_generation2 ^= self.state_generation1
                            simplify_state = self.state.with_keys(self.state_generation2)
                            simplify_time = -system_clock()
                            top = current_program.top.simplify(state=simplify_state, dynamic=dynamic, path=current_program.path)
                            now = system_clock()
                            simplify_time += now
                            if top is current_program.top:
                                logger.trace("Program unchanged after simplification on {} static state keys in {:.1f}ms",
                                             len(self.state_generation2), simplify_time*1000)
                            else:
                                compile_time = -now
                                run_program = top.compile(initial_lnames=tuple(dynamic))
                                run_program.set_path(current_program.path)
                                run_program.set_top(top)
                                compile_time += system_clock()
                                logger.debug("Simplified on {} static state keys to {} instructions in {:.1f}/{:.1f}ms",
                                             len(self.state_generation2), len(run_program), simplify_time*1000, compile_time*1000)
                    self.state_generation1 = self.state_generation0
                    self.state_generation0 = set()
                    simplify_state_time = system_clock() + self.state_simplify_wait

                if self.global_state_dirty and self.state_file is not None and system_clock() > save_state_time:
                    logger.trace("Saving state")
                    with open(self.state_file, 'wb') as file:
                        pickle.dump(self.global_state, file)
                    self.global_state_dirty = False
                    save_state_time = system_clock() + self.STATE_SAVE_PERIOD

                if self.switch_page is not None:
                    self.switch_to_page(self.switch_page)
                    self.switch_page = None
                    run_program = current_program = None
                    performance = 1
                    for renderers in self.renderers.values():
                        for renderer in renderers:
                            await renderer.purge()

                frame_count += 1
                frames.append(frame_time if self.realtime else now)
                frame_period = now - frame_time
                frame_time += 1 / self.target_fps
                if self.realtime:
                    wait_time = frame_time - now
                    performance = min(performance + 0.001, 2) if wait_time > 0.001 else max(0.5, performance - 0.01)
                else:
                    wait_time = 0.001
                if wait_time >= 0.001:
                    slow_frame = False
                    housekeeping -= wait_time - 0.001
                    await asyncio.sleep(wait_time - 0.001)
                else:
                    slow_frame = True
                    logger.trace("Slow frame - {:.0f}ms", frame_period * 1000)
                    await asyncio.sleep(0)
                    frame_time = system_clock()

                now = system_clock()
                housekeeping += now

                if len(frames) > 1 and frames[-1] - frames[0] > 5:
                    nframes = len(frames) - 1
                    fps = nframes / (frames[-1] - frames[0])
                    logger.info("{:4.1f}fps; {:4.1f}/{:4.1f}/{:4.1f}ms (run/render/sys); perf {:.2f}",
                                fps, 1000 * execution / nframes, 1000 * render / nframes, 1000 * housekeeping / nframes, performance)
                    frames = frames[-1:]
                    execution = render = housekeeping = 0
                    logger.trace("State dictionary size: {} keys", len(self.state))
                    if run_program is not None and run_program.stack is not None:
                        logger.trace("VM stack size: {:d}", run_program.stack.size)
                    Model.flush_caches()
                else:
                    gc.collect(0)

        finally:
            self.global_state = {}
            self._references = {}
            self.pages = []
            program = run_program = current_program = context = None
            Model.flush_caches(0, 0)
            SharedCache.clean(0)
            for renderers in self.renderers.values():
                while renderers:
                    await renderers.pop().destroy()
            if self.vm_stats:
                log_vm_stats()
            count = gc.collect(2)
            logger.trace("Collected {} objects (full collection)", count)
            counts = numbers_cache_counts()
            if counts:
                logger.debug("Numbers cache: {}", ", ".join(f"{size}x{count}" for size, count in counts.items()))
            empty_numbers_cache()
            gc.enable()



"""
Flitter render targets
"""

from collections import namedtuple

import av
from loguru import logger
import numpy as np
import PIL.Image

from ...clock import system_clock
from .glconstants import GL_RGBA8, GL_SRGB8_ALPHA8, GL_RGBA16F, GL_RGBA32F, GL_FRAMEBUFFER_SRGB


ColorFormat = namedtuple('ColorFormat', ('moderngl_dtype', 'gl_format', 'srgb_gl_format'))

COLOR_FORMATS = {
    8: ColorFormat('f1', GL_RGBA8, GL_SRGB8_ALPHA8),
    16: ColorFormat('f2', GL_RGBA16F, False),
    32: ColorFormat('f4', GL_RGBA32F, False)
}


class RenderTarget:
    @classmethod
    def get(cls, glctx, width, height, colorbits, *, srgb=False, has_depth=False, samples=0):
        if colorbits not in COLOR_FORMATS:
            colorbits = glctx.extra['colorbits']
        pool = glctx.extra.setdefault('_RenderTarget_pool', {})
        key = width, height, colorbits, srgb, has_depth, samples
        targets = pool.setdefault(key, [])
        if targets:
            target = targets.pop()
            target._release_time = None
            return target
        return RenderTarget(glctx, width, height, colorbits, srgb, has_depth, samples)

    @classmethod
    def empty_pool(cls, glctx, age=0):
        cutoff = system_clock() - age
        pool = glctx.extra.setdefault('_RenderTarget_pool', {})
        for key, targets in pool.items():
            while targets and targets[0]._release_time < cutoff:
                target = targets.pop(0)
                logger.debug("Destroyed {}", str(target))

    def __init__(self, glctx, width, height, colorbits, srgb, has_depth, samples):
        self._glctx = glctx
        self.width = width
        self.height = height
        self.colorbits = colorbits
        self.srgb = srgb
        self.has_depth = has_depth
        self.samples = samples
        self._array = None
        self._image = None
        self._video_frame = None
        self._release_time = None
        format = COLOR_FORMATS[self.colorbits]
        gl_format = format.srgb_gl_format if self.srgb else format.gl_format
        self._image_texture = self._glctx.texture((self.width, self.height), 4, dtype=format.moderngl_dtype, internal_format=gl_format)
        self._depth_renderbuffer = self._glctx.depth_renderbuffer((self.width, self.height), samples=self.samples) if self.has_depth else None
        if self.samples:
            self._color_renderbuffer = self._glctx.renderbuffer((self.width, self.height), 4, samples=self.samples, dtype=format.moderngl_dtype)
            self._render_framebuffer = self._glctx.framebuffer(color_attachments=(self._color_renderbuffer,), depth_attachment=self._depth_renderbuffer)
            self._image_framebuffer = self._glctx.framebuffer(color_attachments=(self._image_texture,))
        else:
            self._color_renderbuffer = None
            self._render_framebuffer = self._glctx.framebuffer(color_attachments=(self._image_texture,), depth_attachment=self._depth_renderbuffer)
            self._image_framebuffer = self._render_framebuffer
        logger.debug("Created {}", str(self))

    def release(self):
        self._release_time = system_clock()
        pool = self._glctx.extra['_RenderTarget_pool']
        key = self.width, self.height, self.colorbits, self.srgb, self.has_depth, self.samples
        pool[key].append(self)

    def __str__(self):
        text = f"{self.width}x{self.height} {self.colorbits}-bit"
        if self.samples:
            text += f" {self.samples}x"
        if self.srgb:
            text += " sRGB"
        text += " render target"
        if self.has_depth:
            text += " with depth"
        return text

    @property
    def size(self):
        return self.width, self.height

    @property
    def texture(self):
        if self._release_time is not None:
            return None
        return self._image_texture

    @property
    def array(self):
        if self._release_time is not None:
            return None
        if self._array is None:
            self._array = np.ndarray((self.height, self.width, 4), 'f4', self._image_framebuffer.read(components=4, dtype='f4'))
        return self._array

    @property
    def image(self):
        if self._release_time is not None:
            return None
        if self._image is None:
            data = self._image_framebuffer.read(components=4, dtype='f1')
            self._image = PIL.Image.frombytes('RGBA', (self.width, self.height), data).transpose(PIL.Image.FLIP_TOP_BOTTOM)
        return self._image

    @property
    def video_frame(self):
        if self._release_time is not None:
            return None
        if self._video_frame is None:
            data = self._image_framebuffer.read(components=4, dtype='f1')
            self._video_frame = av.VideoFrame.from_bytes(data, width=self.width, height=self.height, format='rgba', flip_vertical=True)
        return self._video_frame

    @property
    def framebuffer(self):
        if self._release_time is not None:
            return None
        return self._image_framebuffer

    def clear(self, color=(0, 0, 0, 0)):
        self._render_framebuffer.clear(*tuple(color))

    def use(self):
        if self.srgb:
            self._glctx.enable_direct(GL_FRAMEBUFFER_SRGB)
        self._render_framebuffer.use()

    def finish(self):
        if self.srgb:
            self._glctx.disable_direct(GL_FRAMEBUFFER_SRGB)
        if self._image_framebuffer is not self._render_framebuffer:
            self._glctx.copy_framebuffer(self._image_framebuffer, self._render_framebuffer)
        self._array = None
        self._image = None
        self._video_frame = None

    def depth_write(self, enabled):
        self._render_framebuffer.depth_mask = enabled

    def __enter__(self):
        self.use()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.finish()



"""
Tests of the model.Quaternion class
"""

import math

import numpy as np

from flitter.model import Vector, Quaternion, Matrix44

from . import utils


class TestQuaternion(utils.TestCase):
    def test_construct(self):
        q = Quaternion()
        self.assertEqual(len(q), 4)
        self.assertEqual(q, [1, 0, 0, 0])
        q = Quaternion(0)
        self.assertEqual(len(q), 4)
        self.assertEqual(q, [0, 0, 0, 0])
        q = Quaternion(1)
        self.assertEqual(len(q), 4)
        self.assertEqual(q, [1, 0, 0, 0])
        q = Quaternion(2)
        self.assertEqual(len(q), 4)
        self.assertEqual(q, [2, 0, 0, 0])
        q = Quaternion([1, 2, 3, 4])
        self.assertEqual(len(q), 4)
        self.assertEqual(q, [1, 2, 3, 4])
        self.assertEqual(Quaternion(range(1, 5)), [1, 2, 3, 4])
        self.assertEqual(Quaternion(np.arange(1, 5)), [1, 2, 3, 4])
        with self.assertRaises(ValueError):
            q = Quaternion("Hello world!")
        with self.assertRaises(ValueError):
            q = Quaternion([1, 2, 3])
        with self.assertRaises(ValueError):
            q = Quaternion([1, 2, 3, "four"])
        with self.assertRaises(ValueError):
            q = Quaternion(range(5))
        with self.assertRaises(ValueError):
            q = Quaternion(np.arange(5))

    def test_coerce(self):
        q1 = Quaternion()
        q2 = Quaternion.coerce(q1)
        self.assertIs(q1, q2)
        q1 = [1, 2, 3, 4]
        q2 = Quaternion.coerce(q1)
        self.assertEqual(q1, q2)
        self.assertIsNot(q1, q2)
        q1 = Vector([1, 2, 3, 4])
        q2 = Quaternion.coerce(q1)
        self.assertEqual(q1, q2)
        self.assertIsNot(q1, q2)

    def test_copy(self):
        q1 = Quaternion([1, 2, 3, 4])
        q2 = q1.copy()
        self.assertIsInstance(q2, Quaternion)
        self.assertEqual(q1, q2)
        self.assertIsNot(q1, q2)

    def test_euler(self):
        self.assertEqual(Quaternion.euler([1, 0, 0], 0), [1, 0, 0, 0])
        self.assertEqual(Quaternion.euler([0, 1, 0], 0), [1, 0, 0, 0])
        self.assertEqual(Quaternion.euler([0, 0, 1], 0), [1, 0, 0, 0])
        c, s = math.cos(math.pi*0.25), math.sin(math.pi*0.25)
        self.assertEqual(Quaternion.euler([1, 0, 0], 0.25), [c, s, 0, 0])
        self.assertEqual(Quaternion.euler([0, 1, 0], 0.25), [c, 0, s, 0])
        self.assertEqual(Quaternion.euler([0, 0, 1], 0.25), [c, 0, 0, s])
        self.assertAllAlmostEqual(Quaternion.euler([1, 0, 0], 0.5), [0, 1, 0, 0])
        self.assertAllAlmostEqual(Quaternion.euler([0, 1, 0], 0.5), [0, 0, 1, 0])
        self.assertAllAlmostEqual(Quaternion.euler([0, 0, 1], 0.5), [0, 0, 0, 1])
        c, s = math.cos(math.pi/3), math.sin(math.pi/3)
        self.assertEqual(Quaternion.euler([1, 0, 0], 1/3), [c, s, 0, 0])
        self.assertEqual(Quaternion.euler([0, 1, 0], 1/3), [c, 0, s, 0])
        self.assertEqual(Quaternion.euler([0, 0, 1], 1/3), [c, 0, 0, s])
        self.assertAllAlmostEqual(Quaternion.euler([1, 0, 0], 1), [-1, 0, 0, 0])
        self.assertAllAlmostEqual(Quaternion.euler([0, 1, 0], 1), [-1, 0, 0, 0])
        self.assertAllAlmostEqual(Quaternion.euler([0, 0, 1], 1), [-1, 0, 0, 0])
        self.assertAllAlmostEqual(Quaternion.euler([1, 1, 1], 1/3), [0.5, 0.5, 0.5, 0.5])

    def test_between(self):
        self.assertAllAlmostEqual(Quaternion.between([1, 0, 0], [0, 1, 0]), Quaternion.euler([0, 0, 1], 0.25))
        self.assertAllAlmostEqual(Quaternion.between([0, 0, 1], [1, 0, 0]), Quaternion.euler([0, 1, 0], 0.25))
        self.assertAllAlmostEqual(Quaternion.between([0, 1, 0], [0, 0, 1]), Quaternion.euler([1, 0, 0], 0.25))
        self.assertAllAlmostEqual(Quaternion.between([1, 0, 0], [1, 1, 0]), Quaternion.euler([0, 0, 1], 0.125))
        self.assertAllAlmostEqual(Quaternion.between([0, 0, 1], [1, 0, 1]), Quaternion.euler([0, 1, 0], 0.125))
        self.assertAllAlmostEqual(Quaternion.between([0, 1, 0], [0, 1, 1]), Quaternion.euler([1, 0, 0], 0.125))
        self.assertAllAlmostEqual(Quaternion.between([1, 0, 0], [-1, 0, 0]), Quaternion.euler([0, 1, 0], 0.5))
        self.assertAllAlmostEqual(Quaternion.between([0, 1, 0], [0, -1, 0]), Quaternion.euler([-1, 0, 0], 0.5))
        self.assertAllAlmostEqual(Quaternion.between([0, 0, 1], [0, 0, -1]), Quaternion.euler([0, -1, 0], 0.5))

    def test_multiply(self):
        qx = Quaternion.euler([1, 0, 0], 0.125)
        qy = Quaternion.euler([0, 1, 0], 0.125)
        qz = Quaternion.euler([0, 0, 1], 0.125)
        self.assertAllAlmostEqual(qx @ qx, Quaternion.euler([1, 0, 0], 0.25))
        self.assertAllAlmostEqual(qy @ qy, Quaternion.euler([0, 1, 0], 0.25))
        self.assertAllAlmostEqual(qz @ qz, Quaternion.euler([0, 0, 1], 0.25))
        self.assertAllAlmostEqual(qx @ Quaternion.euler([1, 0, 0], -0.125), Quaternion())
        self.assertAllAlmostEqual(qy @ Quaternion.euler([0, 1, 0], -0.125), Quaternion())
        self.assertAllAlmostEqual(qz @ Quaternion.euler([0, 0, 1], -0.125), Quaternion())
        self.assertNotAllAlmostEqual(qx @ qy, qy @ qx)
        self.assertNotAllAlmostEqual(qx @ qz, qz @ qx)
        self.assertNotAllAlmostEqual(qy @ qz, qz @ qy)
        q = Quaternion.euler([1, 1, 1], 1/3)
        self.assertAllAlmostEqual(q @ q @ q, Quaternion(-1))
        self.assertAllAlmostEqual(q @ q @ q @ q @ q @ q, Quaternion(1))

    def test_inverse(self):
        self.assertAlmostEqual(Quaternion.euler([1, 0, 0], 0.125).inverse(), Quaternion.euler([1, 0, 0], -0.125))
        self.assertAlmostEqual(Quaternion.euler([0, 1, 0], 0.125).inverse(), Quaternion.euler([0, 1, 0], -0.125))
        self.assertAlmostEqual(Quaternion.euler([0, 0, 1], 0.125).inverse(), Quaternion.euler([0, 0, 1], -0.125))
        q = Quaternion.euler([1, 1, 1], 1/3)
        self.assertAlmostEqual(q @ q.inverse(), Quaternion())
        self.assertAlmostEqual(q.inverse() @ q, Quaternion())
        self.assertAlmostEqual(q @ q.inverse() @ q, q)

    def test_normalize(self):
        self.assertAlmostEqual(Quaternion([0.5, 0.5, 0.5, 0.5]).normalize(), [0.5, 0.5, 0.5, 0.5])
        self.assertAlmostEqual(Quaternion([1, 1, 1, 1]).normalize(), [0.5, 0.5, 0.5, 0.5])
        self.assertAlmostEqual(Quaternion(10).normalize(), Quaternion(1))

    def test_conjugate(self):
        qx = Quaternion.euler([1, 0, 0], 0.25)
        self.assertAllAlmostEqual(qx @ [1, 0, 0], [1, 0, 0])
        self.assertAllAlmostEqual(qx @ [0, 1, 0], [0, 0, 1])
        self.assertAllAlmostEqual(qx @ [0, 0, 1], [0, -1, 0])
        qy = Quaternion.euler([0, 1, 0], 0.25)
        self.assertAllAlmostEqual(qy @ [1, 0, 0], [0, 0, -1])
        self.assertAllAlmostEqual(qy @ [0, 1, 0], [0, 1, 0])
        self.assertAllAlmostEqual(qy @ [0, 0, 1], [1, 0, 0])
        qz = Quaternion.euler([0, 0, 1], 0.25)
        self.assertAllAlmostEqual(qz @ [1, 0, 0], [0, 1, 0])
        self.assertAllAlmostEqual(qz @ [0, 1, 0], [-1, 0, 0])
        self.assertAllAlmostEqual(qz @ [0, 0, 1], [0, 0, 1])
        with self.assertRaises(ValueError):
            qx @ None
        with self.assertRaises(ValueError):
            qx @ "Hello"
        with self.assertRaises(ValueError):
            qx @ [1, 0]
        with self.assertRaises(ValueError):
            qx @ [1, 0, 0, 0]

    def test_exponent(self):
        q = Quaternion.euler([1, 1, 1], 1/3)
        i = Quaternion()
        self.assertAllAlmostEqual(q.exponent(-1), q.inverse())
        self.assertAllAlmostEqual(q.exponent(0), i)
        self.assertAllAlmostEqual(q.exponent(1), q)
        self.assertAllAlmostEqual(q.exponent(2), q @ q)
        self.assertAllAlmostEqual(i.exponent(-1), i)
        self.assertAllAlmostEqual(i.exponent(-0.5), i)
        self.assertAllAlmostEqual(i.exponent(0), i)
        self.assertAllAlmostEqual(i.exponent(0.5), i)
        self.assertAllAlmostEqual(i.exponent(1), i)
        self.assertAllAlmostEqual(i.exponent(2), i)

    def test_slerp(self):
        qx = Quaternion.euler([1, 0, 0], 0.25)
        qy = Quaternion.euler([0, 1, 0], 0.25)
        self.assertAllAlmostEqual(qx.slerp(qy, -1), qx @ qy.inverse() @ qx)
        self.assertAllAlmostEqual(qx.slerp(qy, 0), qx)
        self.assertAllAlmostEqual(qx.slerp(qy, 1), qy)
        self.assertAllAlmostEqual(qx.slerp(qy, 2), qy @ qx.inverse() @ qy)
        self.assertAllAlmostEqual(qx.slerp(qx, 0), qx)
        self.assertAllAlmostEqual(qx.slerp(qx, 1), qx)

    def test_matrix44(self):
        self.assertAllAlmostEqual(Quaternion.euler([1, 0, 0], 0.125).matrix44(), Matrix44.rotate_x(0.125))
        self.assertAllAlmostEqual(Quaternion.euler([0, 1, 0], 0.125).matrix44(), Matrix44.rotate_y(0.125))
        self.assertAllAlmostEqual(Quaternion.euler([0, 0, 1], 0.125).matrix44(), Matrix44.rotate_z(0.125))

    def test_repr(self):
        self.assertEqual(repr(Quaternion(0)), "0﹢0𝒊﹢0𝒋﹢0𝒌")
        self.assertEqual(repr(Quaternion(1)), "1﹢0𝒊﹢0𝒋﹢0𝒌")
        self.assertEqual(repr(Quaternion(-1)), "-1﹢0𝒊﹢0𝒋﹢0𝒌")
        self.assertEqual(repr(Quaternion([1, 2, 3, 4])), "1﹢2𝒊﹢3𝒋﹢4𝒌")
        self.assertEqual(repr(Quaternion([-1, -2, -3, -4])), "-1﹣2𝒊﹣3𝒋﹣4𝒌")
        self.assertEqual(repr(Quaternion([1e-10, 2e-10, 3e-10, 4e-10])), "0﹢0𝒊﹢0𝒋﹢0𝒌")
        self.assertEqual(repr(Quaternion([-1e-10, -2e-10, -3e-10, -4e-10])), "0﹢0𝒊﹢0𝒋﹢0𝒌")
